from typing import List, Optional, Tuple

import numpy as np
import pandas as pd
from pandas.api.types import is_integer_dtype, is_categorical
from sklearn.preprocessing import LabelEncoder


def autoprep_gbdt(algorithm_type: str, X_train: pd.DataFrame, X_test: Optional[pd.DataFrame],
                  categorical_feature_to_treat: Optional[List[str]] = None) -> Tuple[pd.DataFrame, pd.DataFrame]:
    if categorical_feature_to_treat is None:
        categorical_feature_to_treat = [c for c in X_train.columns if X_train[c].dtype.name in ['object', 'category']]

    # LightGBM:
    # Can handle categorical dtype. Otherwise, int, float or bool is acceptable for categorical columns.
    # https://lightgbm.readthedocs.io/en/latest/Advanced-Topics.html#categorical-feature-support
    #
    # CatBoost:
    # int, float, bool or str is acceptable for categorical columns. NaN should be filled.
    # https://catboost.ai/docs/concepts/faq.html#why-float-and-nan-values-are-forbidden-for-cat-features
    #
    # XGBoost:
    # All categorical column should be encoded beforehand.

    if algorithm_type == 'lgbm':
        # LightGBM can handle categorical dtype natively
        categorical_feature_to_treat = [c for c in categorical_feature_to_treat if not is_categorical(X_train[c])]

    if algorithm_type == 'cat' and len(categorical_feature_to_treat) > 0:
        X_train = X_train.copy()
        X_test = X_test.copy() if X_test is not None else X_train.iloc[:1, :].copy()  # dummy
        for c in categorical_feature_to_treat:
            X_train[c], X_test[c] = _fill_na_by_unique_value(X_train[c], X_test[c])

    if algorithm_type in ('lgbm', 'xgb') and len(categorical_feature_to_treat) > 0:
        assert X_test is not None, "X_test is required for XGBoost with categorical variables"
        X_train = X_train.copy()
        X_test = X_test.copy()

        for c in categorical_feature_to_treat:
            X_train[c], X_test[c] = _fill_na_by_unique_value(X_train[c], X_test[c])
            le = LabelEncoder()
            concat = np.concatenate([X_train[c].values, X_test[c].values])
            concat = le.fit_transform(concat)
            X_train[c] = concat[:len(X_train)]
            X_test[c] = concat[len(X_train):]

    return X_train, X_test


def _fill_na_by_unique_value(strain: pd.Series, stest: Optional[pd.Series]) -> Tuple[pd.Series, pd.Series]:
    if is_categorical(strain):
        return strain.cat.codes, stest.cat.codes
    elif is_integer_dtype(strain.dtype):
        fillval = min(strain.min(), stest.min()) - 1
        return strain.fillna(fillval), stest.fillna(fillval)
    else:
        return strain.astype(str), stest.astype(str)



parameters = [
    {
        "name": "ieee-2019-1st",
        "url": "https://www.kaggle.com/cdeotte/xgb-fraud-with-magic-0-9600",
        "competition": "ieee-fraud-detection",
        "rank": 1,
        "metric": "auc",
        "parameters": {
            "max_depth": 12,
            "learning_rate": 0.02,
            "subsample": 0.8,
            "colsample_bytree": 0.4,
            "missing": -1,
            "eval_metric": "auc",
            "tree_method": "hist"
        }
    },
    {
        "name": "womens-ml-competition-2019-1st",
        "url": "https://github.com/salmatfq/KaggleMarchMadnessFirstPlace/blob/master/win_ncaa_men.R",
        "competition": "womens-machine-learning-competition-2019",
        "rank": 1,
        "metric": "log-loss",
        "parameters": {
            "eval_metric": "mae",
            "booster": "gbtree",
            "eta": 0.02,
            "subsample": 0.35,
            "colsample_bytree": 0.7,
            "num_parallel_tree": 10,
            "min_child_weight": 40,
            "gamma": 10,
            "max_depth": 3
        }
    },

    # 2018, Santander Value Prediction Challenge
    {
        "name": "santander-2018-5th",
        "url": "https://github.com/vlarine/kaggle/blob/master/santander-value-prediction-challenge/santander.py",
        "competition": "santander-value-prediction-challenge",
        "rank": 5,
        "metric": "rmsle",
        "parameters": {
            'colsample_bytree': 0.055,
            'colsample_bylevel': 0.4,
            'gamma': 1.5,
            'learning_rate': 0.01,
            'max_depth': 5,
            'objective': 'reg:linear',
            'booster': 'gbtree',
            'min_child_weight': 10,
            'reg_alpha': 0,
            'reg_lambda': 0,
            'eval_metric': 'rmse',
            'subsample': 0.7,
        }
    },

    # 2018, Elo Merchant Category Recommendation
    {
        "name": "elo-2018-11th",
        "url": "https://github.com/kangzhang0709/2019-kaggle-elo-top-11-solution/blob/master/Models/model_xgb.ipynb",
        "competition": "elo-merchant-category-recommendation",
        "rank": 11,
        "metric": "rmse",
        "parameters": {
            'objective': 'reg:linear',
            'booster': 'gbtree',
            'learning_rate': 0.01,
            'max_depth': 10,
            'gamma': 1.45,
            'alpha': 0.1,
            'lambda': 0.3,
            'subsample': 0.9,
            'colsample_bytree': 0.054,
            'colsample_bylevel': 0.50
        }
    },

    # 2018, DonorsChoose.org Application Screening
    {
        "name": "donorschoose-2018-1st",
        "url": "https://www.kaggle.com/shadowwarrior/1st-place-solution/notebook",
        "competition": "donorschoose-application-screening",
        "rank": 1,
        "metric": "auc",
        "parameters": {
            'objective': 'binary:logistic',
            'eval_metric': 'auc',
            'eta': 0.01,
            'max_depth': 7,
            'subsample': 0.8,
            'colsample_bytree': 0.4,
            'min_child_weight': 10,
            'gamma': 2
        }
    },

    # 2018, Recruit Restaurant Visitor Forecasting

    # 2017, Instacart Market Basket Analysis
    {
        "name": "instacart-2017-2nd",
        "url": "https://github.com/KazukiOnodera/Instacart/blob/master/py_model/002_xgb_holdout_item_812_1.py",
        "competition": "instacart-market-basket-analysis",
        "rank": 2,
        "metric": "",
        "parameters": {
            'max_depth': 10,
            'eta': 0.02,
            'colsample_bytree': 0.4,
            'subsample': 0.75,
            'eval_metric': 'logloss',
            'objective': 'binary:logistic',
            'tree_method': 'hist'
         }
    },

    # 2017, Two Sigma Connect; Rental Listing Inquiries
    {
        "name": "two-sigma-2017-1st",
        "url": "https://github.com/plantsgo/Rental-Listing-Inquiries/blob/master/xgb.py",
        "competition": "two-sigma-connect-rental-listing-inquiries",
        "rank": 1,
        "metric": "multi-class log-loss",
        "parameters": {
            'booster': 'gbtree',
            'objective': 'multi:softprob',
            'eval_metric': 'mlogloss',
            'gamma': 1,
            'min_child_weight': 1.5,
            'max_depth': 5,
            'lambda': 10,
            'subsample': 0.7,
            'colsample_bytree': 0.7,
            'colsample_bylevel': 0.7,
            'eta': 0.03,
            'tree_method': 'exact'
        }
    },

    # 2016, Santander Product Recommendation
    {
        "name": "santander-2016-2nd",
        "url": "https://github.com/ttvand/Santander-Product-Recommendation/blob/master/First%20level%20learners/xgboost.R",
        "competition": "santander-product-recommendation",
        "rank": 2,
        "metric": "map7",
        "parameters": {
            "etaC": 10,
            "subsample": 1,
            "colsample_bytree": 0.5,
            "max_depth": 8,
            "min_child_weight": 0,
            "gamma": 0.1
        }
    },

    # 2016, TalkingData Mobile User Demographics
    {
        "name": "talkingdata-2016-3rd-1",
        "url": "https://github.com/chechir/talking_data/blob/master/danijel/xgb/xgb_cv5_train_events.R",
        "competition": "talkingdata-mobile-user-demographics",
        "rank": 3,
        "metric": "multi-class log-loss",
        "parameters": {
            "booster": 'gbtree',
            "objective": 'reg:logistic',
            "eval_metric": 'logloss',
            "learning_rate": 0.025,
            "max_depth": 6,
            "subsample": 0.8,
            "colsample_bytree": 0.5,
            "colsample_bylevel": 0.5
        }
    },
    {
        "name": "talkingdata-2016-3rd-2",
        "url": "https://github.com/chechir/talking_data/blob/master/danijel/xgb/xgb_cv5_train_noevents.R",
        "competition": "talkingdata-mobile-user-demographics",
        "rank": 3,
        "metric": "multi-class log-loss",
        "parameters": {
            "booster": 'gbtree',
            "objective": 'reg:logistic',
            "eval_metric": 'logloss',
            "learning_rate": 0.05,
            "max_depth": 2,
            "colsample_bytree": 0.8,
            "colsample_bylevel": 0.8
        }
    },

    # 2016, Allstate Claims Severity
    {
        "name": "allstate-2016-3rd",
        "url": "https://www.kaggle.com/c/allstate-claims-severity/discussion/26447#150319",
        "competition": "allstate-claims-severity",
        "rank": 3,
        "metric": "mae",
        "parameters": {
            'colsample_bytree': 0.4,
            'subsample': 0.975,
            'learning_rate': 0.015,
            'gamma': 1.5,
            'lambda': 2,
            'alpha': 2,
            'max_depth': 25,
            'num_parallel_tree': 1,
            'min_child_weight': 50,
            'eval_metric': 'mae',
            'max_delta_step': 0,
        }
    },

    # 2016, Bosch Production Line Performance
    {
        "name": "bosch-2016-1st",
        "url": "https://www.kaggle.com/c/bosch-production-line-performance/discussion/25434#144628",
        "competition": "bosch-production-line-performance",
        "rank": 1,
        "metric": "mcc",
        "parameters": {
            "eval_metric": "auc",
            "alpha": 0,
            "booster": "gbtree",
            "colsample_bytree": 0.6,
            "minchildweight": 5,
            "subsample": 0.9,
            "eta": 0.03,
            "objective": "binary:logistic",
            "max_depth": 14,
            "lambda": 4
        }
    },
]



from typing import Callable, List, Optional, Tuple

import numpy as np
import pandas as pd
import scipy.stats as stats
from scipy.optimize import minimize

from nyaggle.ensemble.common import EnsembleResult


def averaging(test_predictions: List[np.ndarray],
              oof_predictions: Optional[List[np.ndarray]] = None,
              y: Optional[pd.Series] = None,
              weights: Optional[List[float]] = None,
              eval_func: Optional[Callable] = None,
              rank_averaging: bool = False) -> EnsembleResult:
    """
    Perform averaging on model predictions.

    Args:
        test_predictions:
            List of predicted values on test data.
        oof_predictions:
            List of predicted values on out-of-fold training data.
        y:
            Target value
        weights:
            Weights for each predictions
        eval_func:
            Evaluation metric used for calculating result score. Used only if ``oof_predictions`` and ``y`` are given.
        rank_averaging:
            If ``True``, predictions will be converted to rank before averaging.
    Returns:
        Namedtuple with following members

        * test_prediction:
            numpy array, Average prediction on test data.
        * oof_prediction:
            numpy array, Average prediction on Out-of-Fold validation data. ``None`` if ``oof_predictions`` = ``None``.
        * score:
            float, Calculated score on Out-of-Fold data. ``None`` if ``eval_func`` is ``None``.
    """
    if weights is None:
        weights = np.ones((len(test_predictions))) / len(test_predictions)

    if rank_averaging:
        test_predictions, oof_predictions = _to_rank(test_predictions, oof_predictions)

    def _weighted_average(predictions: List[np.ndarray], weights: List[float]):
        if len(predictions) != len(weights):
            raise ValueError('len(predictions) != len(weights)')
        average = np.zeros_like(predictions[0])

        for i, weight in enumerate(weights):
            if predictions[i].shape != average.shape:
                raise ValueError('predictions[{}].shape != predictions[0].shape'.format(i))
            average += predictions[i] * weight

        return average

    average_test = _weighted_average(test_predictions, weights)
    if oof_predictions is not None:
        average_oof = _weighted_average(oof_predictions, weights)
        score = eval_func(y, average_oof) if eval_func is not None else None
    else:
        average_oof = None
        score = None

    return EnsembleResult(average_test, average_oof, score)


def averaging_opt(test_predictions: List[np.ndarray],
                  oof_predictions: Optional[List[np.ndarray]],
                  y: Optional[pd.Series],
                  eval_func: Optional[Callable],
                  higher_is_better: bool,
                  weight_bounds: Tuple = (0, 1),
                  rank_averaging: bool = False) -> EnsembleResult:
    """
    Perform averaging with optimal weights using scipy.optimize

    Args:
        test_predictions:
            List of predicted values on test data.
        oof_predictions:
            List of predicted values on out-of-fold training data.
        y:
            Target value
        eval_func:
            Evaluation metric used for calculating result score. Used only if ``oof_predictions`` and ``y`` are given.
        higher_is_better:
            Determine the direction of optimize ``eval_func``.
        weight_bounds:
            Specify lower/upper bounds of each weight.
        rank_averaging:
            If ``True``, predictions will be converted to rank before averaging.
    Returns:
        Namedtuple with following members

        * test_prediction:
            numpy array, Average prediction on test data.
        * oof_prediction:
            numpy array, Average prediction on Out-of-Fold validation data. ``None`` if ``oof_predictions`` = ``None``.
        * score:
            float, Calculated score on Out-of-Fold data. ``None`` if ``eval_func`` is ``None``.
    """

    def _minimize(weights):
        prediction = np.zeros_like(oof_predictions[0])
        for weight, oof in zip(weights, oof_predictions):
            prediction += weight * oof
        oof_score = eval_func(y, prediction)

        return -oof_score if higher_is_better else oof_score

    weights = np.ones((len(test_predictions))) / len(test_predictions)

    if rank_averaging:
        test_predictions, oof_predictions = _to_rank(test_predictions, oof_predictions)

    cons = ({'type': 'eq', 'fun': lambda w: 1 - sum(w)})

    bounds = [weight_bounds] * len(test_predictions)

    result = minimize(_minimize, weights, method='SLSQP', constraints=cons, bounds=bounds)

    return averaging(test_predictions, oof_predictions, y, result['x'], eval_func)


def _to_rank(test_predictions: List[np.ndarray], oof_predictions: Optional[List[np.ndarray]]):
    if oof_predictions is not None:
        oof_predictions = [stats.rankdata(oof) / len(oof) for oof in oof_predictions]
    test_predictions = [stats.rankdata(test) / len(test) for test in test_predictions]

    return test_predictions, oof_predictions



import os
import shutil
import tempfile
import uuid
from contextlib import contextmanager
from typing import Tuple

import numpy as np
import pandas as pd
from sklearn.datasets import make_classification, make_regression


def make_classification_df(n_samples: int = 1024,
                           n_num_features: int = 20,
                           n_cat_features: int = 0,
                           class_sep: float = 1.0,
                           n_classes: int = 2,
                           feature_name: str = 'col_{}',
                           target_name: str = 'target',
                           random_state: int = 0,
                           id_column: str = None) -> Tuple[pd.DataFrame, pd.Series]:
    np.random.seed(random_state)
    X, y = make_classification(n_samples=n_samples, n_features=n_num_features, class_sep=class_sep,
                               random_state=random_state, n_classes=n_classes, n_informative=max(n_classes, 2))

    X = pd.DataFrame(X, columns=[feature_name.format(i) for i in range(n_num_features)])
    y = pd.Series(y, name=target_name)

    if id_column is not None:
        X[id_column] = range(n_samples)

    for i in range(n_cat_features):
        X['cat_{}'.format(i)] = \
            pd.Series(np.random.choice(['A', 'B', None], size=n_samples)).astype('category')

    return X, y


def make_regression_df(n_samples: int = 1024,
                       n_num_features: int = 20,
                       n_cat_features: int = 0,
                       feature_name: str = 'col_{}',
                       target_name: str = 'target',
                       random_state: int = 0,
                       id_column: str = None) -> Tuple[pd.DataFrame, pd.Series]:
    np.random.seed(random_state)
    X, y = make_regression(n_samples=n_samples, n_features=n_num_features,
                           random_state=random_state)

    X = pd.DataFrame(X, columns=[feature_name.format(i) for i in range(n_num_features)])
    y = pd.Series(y, name=target_name)

    if id_column is not None:
        X[id_column] = range(n_samples)

    for i in range(n_cat_features):
        X['cat_{}'.format(i)] = \
            pd.Series(np.random.choice(['A', 'B', None], size=n_samples)).astype(str).astype('category')

    return X, y




@contextmanager
def get_temp_directory() -> str:
    path = None
    try:
        path = os.path.join(tempfile.gettempdir(), uuid.uuid4().hex)
        yield path
    finally:
        if path:
            shutil.rmtree(path, ignore_errors=True)




# Original work of safe_instance:
# https://github.com/slundberg/shap/blob/master/shap/common.py
# -----------------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2018 Scott Lundberg
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -----------------------------------------------------------------------------

import importlib
from typing import List, Tuple, Union


def is_instance(obj, class_path_str: Union[str, List, Tuple]) -> bool:
    """
    Acts as a safe version of isinstance without having to explicitly
    import packages which may not exist in the users environment.
    Checks if obj is an instance of type specified by class_path_str.
    Parameters
    ----------
    obj: Any
        Some object you want to test against
    class_path_str: str or list
        A string or list of strings specifying full class paths
        Example: `sklearn.ensemble.RandomForestRegressor`
    Returns
    --------
    bool: True if isinstance is true and the package exists, False otherwise
    """
    if isinstance(class_path_str, str):
        class_path_strs = [class_path_str]
    elif isinstance(class_path_str, list) or isinstance(class_path_str, tuple):
        class_path_strs = class_path_str
    else:
        class_path_strs = ['']

    # try each module path in order
    for class_path_str in class_path_strs:
        if "." not in class_path_str:
            raise ValueError("class_path_str must be a string or list of strings specifying a full \
                module path to a class. Eg, 'sklearn.ensemble.RandomForestRegressor'")

        # Splits on last occurence of "."
        module_name, class_name = class_path_str.rsplit(".", 1)

        # Check module exists
        try:
            spec = importlib.util.find_spec(module_name)
        except:
            spec = None
        if spec is None:
            continue

        module = importlib.import_module(module_name)

        # Get class
        _class = getattr(module, class_name, None)
        if _class is None:
            continue

        if isinstance(obj, _class):
            return True

    return False


def is_gbdt_instance(obj, algorithm_type: Union[str, Tuple]) -> bool:
    if isinstance(algorithm_type, str):
        algorithm_type = (algorithm_type,)

    gbdt_instance_name = {
        'lgbm': 'lightgbm.sklearn.LGBMModel',
        'xgb': 'xgboost.sklearn.XGBModel',
        'cat': 'catboost.core.CatBoost'
    }

    return is_instance(obj, tuple(gbdt_instance_name[t] for t in algorithm_type))



import os
import shutil
import tempfile
import uuid

import pytest


@pytest.fixture(scope='function', autouse=True)
def tmpdir_name():
    path = None
    try:
        path = os.path.join(tempfile.gettempdir(), uuid.uuid4().hex)
        yield path
    finally:
        if path:
            shutil.rmtree(path, ignore_errors=True)



from typing import Optional, Tuple

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns


def plot_importance(importance: pd.DataFrame, path: Optional[str] = None, top_n: int = 100,
                    figsize: Optional[Tuple[int, int]] = None,
                    title: Optional[str] = None):
    """
    Plot feature importance and write to image

    Args:
        importance:
            The dataframe which has "feature" and "importance" column
        path:
            The file path to be saved
        top_n:
            The number of features to be visualized
        figsize:
            The size of the figure
        title:
            The title of the plot
    Example:
        >>> import pandas as pd
        >>> import lightgbm as lgb
        >>> from nyaggle.util import plot_importance
        >>> from sklearn.datasets import make_classification

        >>> X, y = make_classification()
        >>> X = pd.DataFrame(X, columns=['col{}'.format(i) for i in range(X.shape[1])])
        >>> booster = lgb.train({'objective': 'binary'}, lgb.Dataset(X, y))
        >>> importance = pd.DataFrame({
        >>>     'feature': X.columns,
        >>>     'importance': booster.feature_importance('gain')
        >>> })
        >>> plot_importance(importance, 'importance.png')
    """
    importance = importance.groupby('feature')['importance'] \
        .mean() \
        .reset_index() \
        .sort_values(by='importance', ascending=False)

    if len(importance) > top_n:
        importance = importance.iloc[:top_n, :]

    if figsize is None:
        figsize = (10, 16)

    if title is None:
        title = 'Feature Importance'

    plt.figure(figsize=figsize)
    sns.barplot(x="importance", y="feature", data=importance)
    plt.title(title)
    plt.tight_layout()
    if path is not None:
        plt.savefig(path)



from typing import Any, Callable, List, Optional, Union
import transformers

import numpy as np
import pandas as pd
from category_encoders.utils import convert_input
from sklearn.decomposition import TruncatedSVD
from tqdm import tqdm

from nyaggle.environment import requires_torch
from nyaggle.feature.base import BaseFeaturizer


class BertSentenceVectorizer(BaseFeaturizer):
    """Sentence Vectorizer using BERT pretrained model.

    Extract fixed-length feature vector from English/Japanese variable-length sentence using BERT.

    Args:
        lang:
            Type of language. If set to "jp", Japanese BERT model is used (you need to install MeCab).
        n_components:
            Number of components in SVD. If `None`, SVD is not applied.
        text_columns:
            List of processing columns. If `None`, all object columns are regarded as text column.
        pooling_strategy:
            The pooling algorithm for generating fixed length encoding vector. 'reduce_mean' and 'reduce_max' use
            average pooling and max pooling respectively to reduce vector from (num-words, emb-dim) to (emb_dim).
            'reduce_mean_max' performs 'reduce_mean' and 'reduce_max' separately and concat them.
            'cls_token' takes the first element (i.e. [CLS]).
        use_cuda:
            If `True`, inference is performed on GPU.
        tokenizer:
            The custom tokenizer used instead of default tokenizer
        model:
            The custom pretrained model used instead of default BERT model
        return_same_type:
            If True, `transform` and `fit_transform` return the same type as X.
            If False, these APIs always return a numpy array, similar to sklearn's API.
        column_format:
            Name of transformed columns (used if returning type is pd.DataFrame)
    """

    def __init__(self, lang: str = 'en', n_components: Optional[int] = None,
                 text_columns: List[str] = None, pooling_strategy: str = 'reduce_mean',
                 use_cuda: bool = False, tokenizer: transformers.PreTrainedTokenizer = None,
                 model=None, return_same_type: bool = True, column_format: str = '{col}_{idx}'):
        if tokenizer is not None:
            assert model is not None
            self.tokenizer = tokenizer
            self.model = model
        if lang == 'en':
            pretrained_model_name = 'bert-base-uncased'
            self.tokenizer = transformers.BertTokenizer.from_pretrained(pretrained_model_name)
            self.model = transformers.BertModel.from_pretrained(pretrained_model_name)
        elif lang == 'jp':
            pretrained_model_name = 'cl-tohoku/bert-base-japanese-whole-word-masking'
            self.tokenizer = transformers.BertJapaneseTokenizer.from_pretrained(pretrained_model_name)
            self.model = transformers.BertModel.from_pretrained(pretrained_model_name)
        else:
            raise ValueError('Specified language type () is invalid.'.format(lang))

        self.lang = lang
        self.n_components = n_components
        self.text_columns = text_columns
        self.pooling_strategy = pooling_strategy
        self.use_cuda = use_cuda
        self.return_same_type = return_same_type
        self.svd = {}
        self.column_format = column_format

    def _process_text(self, text: str) -> np.ndarray:
        requires_torch()
        import torch

        tokens_tensor = torch.tensor([self.tokenizer.encode(text, add_special_tokens=True)])
        if self.use_cuda:
            tokens_tensor = tokens_tensor.to('cuda')
            self.model.to('cuda')

        self.model.eval()
        with torch.no_grad():
            outputs = self.model(tokens_tensor)

        embedding = outputs.last_hidden_state.cpu().numpy()[0]
        if self.pooling_strategy == 'reduce_mean':
            return np.mean(embedding, axis=0)
        elif self.pooling_strategy == 'reduce_max':
            return np.max(embedding, axis=0)
        elif self.pooling_strategy == 'reduce_mean_max':
            return np.r_[np.max(embedding, axis=0), np.mean(embedding, axis=0)]
        elif self.pooling_strategy == 'cls_token':
            return embedding[0]
        else:
            raise ValueError("specify valid pooling_strategy: {reduce_mean, reduce_max, reduce_mean_max, cls_token}")

    def _fit_one(self, col: str, emb: np.ndarray):
        if not self.n_components or self.n_components >= emb.shape[1]:
            return emb
        self.svd[col] = TruncatedSVD(n_components=self.n_components, algorithm='arpack', random_state=0)
        return self.svd[col].fit(emb)

    def _transform_one(self, col: str, emb: np.ndarray):
        if not self.n_components or self.n_components >= emb.shape[1]:
            return emb
        return self.svd[col].transform(emb)

    def _fit_transform_one(self, col: str, emb: np.ndarray):
        if not self.n_components or self.n_components >= emb.shape[1]:
            return emb
        self.svd[col] = TruncatedSVD(n_components=self.n_components, algorithm='arpack', random_state=0)
        return self.svd[col].fit_transform(emb)

    def _process(self, X: pd.DataFrame, func: Callable[[str, np.ndarray], Any]):
        is_pandas = isinstance(X, pd.DataFrame)
        X = convert_input(X)

        tqdm.pandas()
        columns = self.text_columns or [c for c in X.columns if X[c].dtype == np.object]
        non_text_columns = [c for c in X.columns if c not in columns]

        column_names = []
        processed = []
        for c in columns:
            emb = np.vstack(X[c].progress_apply(lambda x: self._process_text(x)))
            emb = func(c, emb)
            processed.append(emb)
            column_names += [self.column_format.format(col=c, idx=i) for i in range(emb.shape[1])]

        processed_df = pd.DataFrame(np.hstack(processed), columns=column_names)

        if non_text_columns:
            X_ = X[non_text_columns].copy()
            X_ = pd.concat([X_, processed_df], axis=1)
        else:
            X_ = processed_df

        return X_ if self.return_same_type and is_pandas else X_.values

    def fit(self, X: Union[pd.DataFrame, np.ndarray], y=None):
        """
        Fit SVD model on training data X.

        Args:
            X:
                Data
            y:
                Ignored
        """
        self._process(X, self._fit_one)
        return self

    def transform(self, X: Union[pd.DataFrame, np.ndarray], y=None):
        """
        Perform feature extraction and dimensionality reduction using
        BERT pre-trained model and trained SVD model.

        Args:
            X:
                Data
            y:
                Ignored
        """
        return self._process(X, self._transform_one)

    def fit_transform(self, X: Union[pd.DataFrame, np.ndarray], y=None, **fit_params):
        """
        Fit SVD model on training data X and perform feature extraction and dimensionality reduction using
        BERT pre-trained model and trained SVD model.

        Args:
            X:
                Data
            y:
                Ignored
        """
        return self._process(X, self._fit_transform_one)



# seed, verbosity, num_boosting_rounds, num_class, num_thread are removed from original code
# (too specific to the competition or environment)

parameters = [
    # 2019, ASHRAE - Great Energy Predictor III

    # 2019, IEEE-CIS Fraud Detection
    {
        "name": "ieee-2019-10th",
        "url": "https://github.com/jxzly/Kaggle-IEEE-CIS-Fraud-Detection-2019/blob/master/main.py",
        "competition": "ieee-fraud-detection",
        "rank": 10,
        "metric": "auc",
        "parameters": {
            'num_leaves': 333,
            'min_child_weight': 0.03454472573214212,
            'feature_fraction': 0.3797454081646243,
            'bagging_fraction': 0.4181193142567742,
            'min_data_in_leaf': 106,
            'objective': 'binary',
            'max_depth': -1,
            'learning_rate': 0.006883242363721497,
            "boosting_type": "gbdt",
            "metric": 'auc',
            'reg_alpha': 0.3899927210061127,
            'reg_lambda': 0.6485237330340494
        }
    },
    {
        "name": "ieee-2019-17th",
        "url": "https://nbviewer.jupyter.org/github/tmheo/IEEE-Fraud-Detection-17th-Place-Solution/blob/master/notebook/IEEE-17th-Place-Solution-LightGBM.ipynb",
        "competition": "ieee-fraud-detection",
        "rank": 17,
        "metric": "auc",
        "parameters": {
            'objective': 'binary',
            'metric': 'None',
            'learning_rate': 0.01,
            'num_leaves': 256,
            'max_bin': 255,
            'max_depth': -1,
            'bagging_freq': 5,
            'bagging_fraction': 0.7,
            'feature_fraction': 0.7,
            'first_metric_only': True,
        }
    },

    # 2019, Predicting Molecular Properties

    # 2019, Instant Gratification

    # 2019, LANL Earthquake Prediction
    {
        "name": "lanl-2019-1st",
        "url": "https://www.kaggle.com/ilu000/1-private-lb-kernel-lanl-lgbm/",
        "competition": "LANL-Earthquake-Prediction",
        "rank": 1,
        "metric": "mae",
        "parameters": {
            'num_leaves': 4,
            'min_data_in_leaf': 5,
            'objective': 'fair',
            'max_depth': -1,
            'learning_rate': 0.02,
            "boosting": "gbdt",
            'boost_from_average': True,
            "feature_fraction": 0.9,
            "bagging_freq": 1,
            "bagging_fraction": 0.5,
            'max_bin': 500,
            'reg_alpha': 0,
            'reg_lambda': 0
        }
    },
    {
        "name": "lanl-2019-10th",
        "url": "https://www.kaggle.com/c/LANL-Earthquake-Prediction/discussion/94466",
        "competition": "LANL-Earthquake-Prediction",
        "rank": 10,
        "metric": "mae",
        "parameters": {
            "objective": "gamma",
            "max_depth": -1,
            "feature_fraction": 0.025,
            "bagging_fraction": 0.250,
            "bagging_freq": 1,
            "num_leaves": 7,
            "min_data_in_bin": 2,
            "max_bin": 25,
            "min_data_in_leaf": 4,
            "lambda_l1": 1.1,
            "lambda_l2": 0.1,
            "learning_rate": 0.01
        }
    },

    # 2019, Santander Customer Transaction Prediction
    {
        "name": "santander-2019-2nd",
        "url": "https://github.com/KazukiOnodera/santander-customer-transaction-prediction/blob/master/py/990_2nd_place_solution_golf.py",
        "competition": "santander-customer-transaction-prediction",
        "rank": 2,
        "metric": "auc",
        "parameters": {
            'bagging_freq': 5,
            'bagging_fraction': 1.0,
            'boost_from_average': 'false',
            'boost': 'gbdt',
            'feature_fraction': 1.0,
            'learning_rate': 0.005,
            'max_depth': -1,
            'min_data_in_leaf': 30,
            'min_sum_hessian_in_leaf': 10.0,
            'num_leaves': 64,
            'tree_learner': 'serial',
            'metric': 'binary_logloss',
            'objective': 'binary',
        }
    },
    {
        "name": "santander-2019-5th",
        "url": "https://github.com/tnmichael309/Kaggle-Santander-Customer-Transaction-Prediction-5th-Place-Partial-Solution/blob/master/notebooks/LGB%20Model.ipynb",
        "competition": "santander-customer-transaction-prediction",
        "rank": 5,
        "metric": "auc",
        "parameters": {
            'num_leaves': 8,
            'min_data_in_leaf': 42,
            'objective': 'binary',
            'max_depth': 16,
            'learning_rate': 0.03,
            'boosting': 'gbdt',
            'bagging_freq': 5,
            'bagging_fraction': 0.8,
            'feature_fraction': 0.8201,
            'reg_alpha': 1.7289,
            'reg_lambda': 4.984,
            'metric': 'auc',
            'subsample': 0.81,
            'min_gain_to_split': 0.01,
            'min_child_weight': 19.428
        }
    },

    # 2019, PetFinder.my Adoption Prediction
    {
        "name": "petfinder-2019-2nd",
        "url": "https://github.com/okotaku/pet_finder/blob/master/code/stack-480-speedup.py",
        "competition": "petfinder-adoption-prediction",
        "rank": 2,
        "metric": "qwk",
        "parameters": {
            'objective': 'regression',
            'metric': 'rmse',
            'learning_rate': 0.01,
            'num_leaves': 63,
            'subsample': 0.9,
            'subsample_freq': 1,
            'colsample_bytree': 0.6,
            'max_depth': 9,
            'max_bin': 127,
            'reg_alpha': 0.11,
            'reg_lambda': 0.01,
            'min_child_weight': 0.2,
            'min_child_samples': 20,
            'min_gain_to_split': 0.02,
            'min_data_in_bin': 3,
            'bin_construct_sample_cnt': 5000,
            'cat_l2': 10
        }
    },

    # 2018, VSB Power Line Fault Detection
    {
        "name": "vsb-2018-1st",
        "url": "https://www.kaggle.com/mark4h/vsb-1st-place-solution",
        "competition": "vsb-power-line-fault-detection",
        "rank": 1,
        "metric": "mcc",
        "parameters": {
            'objective': 'binary',
            'boosting': 'gbdt',
            'metric': 'binary_logloss',
            'learning_rate': 0.01,
            'num_leaves': 80,
            'feature_fraction': 0.8,
            'bagging_freq': 1,
            'bagging_fraction': 0.8
        }
    },

    # 2018, Microsoft Malware Prediction
    {
        "name": "microsoft-2018-2nd",
        "url": "https://github.com/imor-de/microsoft_malware_prediction_kaggle_2nd/blob/master/code/7_Submission_M2.ipynb",
        "competition": "microsoft-malware-prediction",
        "rank": 2,
        "metric": "auc",
        "parameters": {
            'boosting_type': 'gbdt',
            'colsample_bytree': 0.6027132059774907,
            'learning_rate': 0.010899921631042043,
            'min_child_samples': 145,
            'num_leaves': 156,
            'reg_alpha': 0.45996805852518485,
            'reg_lambda': 0.7336912016500579,
            'subsample_for_bin': 440000,
            'subsample': 0.5512957111882841
        }
    },
    {
        "name": "microsoft-2018-4th",
        "url": "https://github.com/Johnnyd113/Microsoft-Malware-Prediction/blob/master/train.py",
        "competition": "microsoft-malware-prediction",
        "rank": 4,
        "metric": "auc",
        "parameters": {
            'num_leaves': 128,
            'min_data_in_leaf': 42,
            'objective': 'binary',
            'metric': 'auc',
            'max_depth': -1,
            'learning_rate': 0.05,
            "boosting": "gbdt",
            "feature_fraction": 0.8,
            "bagging_freq": 5,
            "bagging_fraction": 0.8,
            "lambda_l1": 0.15,
            "lambda_l2": 0.15
        }
    },

    # 2018, Elo Merchant Category Recommendation
    {
        "name": "elo-2018-11th",
        "url": "https://github.com/kangzhang0709/2019-kaggle-elo-top-11-solution/blob/master/Models/model_lgbm.ipynb",
        "competition": "elo-merchant-category-recommendation",
        "rank": 11,
        "metric": "rmse",
        "parameters": {
            'objective': 'regression',
            'metric': 'rmse',
            'learning_rate': 0.01,
            'subsample': 0.78,
            'max_depth': 8,
            'num_leaves': 63,
            'min_child_weight': 41.9612,
            'reg_alpha': 9.677,
            'colsample_bytree': 0.566,
            'min_split_gain': 8.820,
            'reg_lambda': 9.253,
            'min_data_in_leaf': 21,
        }
    },

    # 2018, Google Analytics Customer Revenue Prediction
    {
        "name": "ga-2018-1st",
        "url": "https://www.kaggle.com/kostoglot/winning-solution",
        "competition": "ga-customer-revenue-prediction",
        "rank": 1,
        "metric": "rmse",
        "parameters": {
            "objective": "regression",
            "max_bin": 256,
            "learning_rate": 0.01,
            "num_leaves": 9,
            "bagging_fraction": 0.9,
            "feature_fraction": 0.8,
            "min_data": 1,
            "bagging_freq": 1,
            "metric": "rmse"
        }
    },

    # 2018, PLAsTiCC Astronomical Classification
    {
        "name": "plasticc-2018-1st",
        "url": "https://github.com/kboone/avocado/blob/master/avocado/classifier.py",
        "competition": "PLAsTiCC-2018",
        "rank": 1,
        "metric": "multi-class log-loss",
        "parameters": {
            "boosting_type": "gbdt",
            "objective": "multiclass",
            "metric": "multi_logloss",
            "learning_rate": 0.05,
            "colsample_bytree": 0.5,
            "reg_alpha": 0.0,
            "reg_lambda": 0.0,
            "min_split_gain": 10.0,
            "min_child_weight": 2000.0,
            "max_depth": 7,
            "num_leaves": 50,
        }
    },
    {
        "name": "plasticc-2018-3rd",
        "url": "https://github.com/nyanp/kaggle-PLASTiCC/blob/master/model/experiment65.py",
        "competition": "PLAsTiCC-2018",
        "rank": 3,
        "metric": "multi-class log-loss",
        "parameters": {
            'boosting_type': 'gbdt',
            'objective': 'multiclass',
            'metric': 'multi_logloss',
            'subsample': .9,
            'colsample_bytree': .9,
            'reg_alpha': 0,
            'reg_lambda': 3,
            'min_split_gain': 0,
            'min_child_weight': 10,
            'learning_rate': 0.1,
            'max_depth': 3,
            'min_data_in_leaf': 1,
            'max_bin': 128,
            'bagging_fraction': 0.66
        }
    },

    # 2018, Home Credit Default Risk
    {
        "name": "home-credit-2018-2nd",
        "url": "https://github.com/KazukiOnodera/Home-Credit-Default-Risk/blob/master/py/935_predict_908-2.py",
        "competition": "home-credit-default-risk",
        "rank": 2,
        "metric": "auc",
        "parameters": {
            'objective': 'binary',
            'metric': 'auc',
            'learning_rate': 0.01,
            'max_depth': 6,
            'num_leaves': 63,
            'max_bin': 255,
            'min_child_weight': 10,
            'min_data_in_leaf': 150,
            'reg_lambda': 0.5,
            'reg_alpha': 0.5,
            'colsample_bytree': 0.7,
            'subsample': 0.9,
            'bagging_freq': 1,
        }
    },
    {
        "name": "home-credit-2018-7th",
        "url": "https://github.com/js-aguiar/home-credit-default-competition/blob/master/config.py",
        "competition": "home-credit-default-risk",
        "rank": 7,
        "metric": "auc",
        "parameters": {
            'boosting_type': 'goss',
            'learning_rate': 0.005134,
            'num_leaves': 54,
            'max_depth': 10,
            'subsample_for_bin': 240000,
            'reg_alpha': 0.436193,
            'reg_lambda': 0.479169,
            'colsample_bytree': 0.508716,
            'min_split_gain': 0.024766,
            'subsample': 1
        }
    },

    # 2018, Santander Value Prediction Challenge
    {
        "name": "santander-2018-6th",
        "url": "https://www.kaggle.com/joeytaj/leak-fe-ml-from-scratch-baseline/notebook",
        "competition": "santander-value-prediction-challenge",
        "rank": 6,
        "metric": "rmsle",
        "parameters": {
            'objective': 'regression',
            'metric': 'rmse',
            'boosting': 'gbdt',
            'is_training_metric': True,
            'max_bin': 350,
            'learning_rate': .005,
            'max_depth': -1,
            'num_leaves': 48,
            'feature_fraction': 0.1,
            'reg_alpha': 0,
            'reg_lambda': 0.2,
            'min_child_weight': 10
        }
    },

    # 2018, Avito Demand Prediction Challenge
    {
        "name": "avito-2018-4th",
        "url": "https://www.kaggle.com/c/avito-demand-prediction/discussion/59881",
        "competition": "avito-demand-prediction",
        "rank": 4,
        "metric": "rmse",
        "parameters": {
            'boosting_type': 'gbdt',
            'objective': 'regression',
            'metric': 'rmse',
            'learning_rate': 0.01,
            'num_leaves': 400,
            'colsample_bytree': .45
        }
    },
    {
        "name": "avito-2018-5th",
        "url": "https://github.com/darraghdog/avito-demand/blob/master/lgb/lb_1406.py",
        "competition": "avito-demand-prediction",
        "rank": 5,
        "metric": "rmse",
        "parameters": {
            'task': 'train',
            'boosting_type': 'gbdt',
            'objective': 'regression',
            'metric': 'rmse',
            'num_leaves': 1000,
            'learning_rate': 0.02,
            'feature_fraction': 0.5,
        }
    },

    # 2018, TalkingData AdTracking Fraud Detection Challenge
    {
        "name": "talkingdata-2018-1st",
        "url": "https://github.com/flowlight0/talkingdata-adtracking-fraud-detection/blob/master/configs/lightgbm_119.json",
        "competition": "talkingdata-adtracking-fraud-detection",
        "rank": 1,
        "metric": "auc",
        "parameters": {
            "boosting_type": "gbdt",
            "objective": "binary",
            "learning_rate": 0.01,
            "num_leaves": 255,
            "max_depth": 8,
            "min_child_samples": 200,
            "subsample": 0.9,
            "subsample_freq": 1,
            "colsample_bytree": 0.5,
            "min_child_weight": 0,
            "subsample_for_bin": 1000000,
            "min_split_gain": 0,
            "reg_lambda": 0,
        }
    },
    {
        "name": "porto-seguro-2017-2nd",
        "url": "https://www.kaggle.com/xiaozhouwang/2nd-place-lightgbm-solution",
        "competition": "porto-seguro-safe-driver-prediction",
        "rank": 2,
        "metric": "normalized-gini-coefficient",
        "parameters": {
            "objective": "binary",
            "learning_rate": 0.1,
            "num_leaves": 15,
            "max_bin": 256,
            "feature_fraction": 0.6,
            "drop_rate": 0.1,
            "is_unbalance": False,
            "max_drop": 50,
            "min_child_samples": 10,
            "min_child_weight": 150,
            "min_split_gain": 0,
            "subsample": 0.9
        }
    },

    # 2018, DonorsChoose.org Application Screening
    {
        "name": "donorschoose-2018-1st",
        "url": "https://www.kaggle.com/shadowwarrior/1st-place-solution/notebook",
        "competition": "donorschoose-application-screening",
        "rank": 1,
        "metric": "auc",
        "parameters": {
            'objective': 'binary',
            'metric': 'auc',
            'boosting_type': 'dart',
            'learning_rate': 0.01,
            'max_bin': 15,
            'max_depth': 17,
            'num_leaves': 63,
            'subsample': 0.8,
            'subsample_freq': 5,
            'colsample_bytree': 0.8,
            'reg_lambda': 7
        }
    },

    # 2018, Recruit Restaurant Visitor Forecasting
    {
        "name": "recruit-2018-1st",
        "url": "https://www.kaggle.com/plantsgo/solution-public-0-471-private-0-505",
        "competition": "recruit-restaurant-visitor-forecasting",
        "rank": 1,
        "metric": "rmse",
        "parameters": {
            'num_leaves': 255,
            'objective': 'regression_l2',
            'max_depth': 9,
            'min_data_in_leaf': 50,
            'learning_rate': 0.007,
            'feature_fraction': 0.6,
            'bagging_fraction': 0.8,
            'bagging_freq': 1,
            'metric': 'rmse'
        }
    },
    {
        "name": "recruit-2018-8th",
        "url": "https://github.com/MaxHalford/kaggle-recruit-restaurant/blob/master/Solution.ipynb",
        "competition": "recruit-restaurant-visitor-forecasting",
        "rank": 8,
        "metric": "rmse",
        "parameters": {
            "objective": 'regression',
            "max_depth": 5,
            "num_leaves": 24,
            "learning_rate": 0.007,
            "n_estimators": 30000,
            "min_child_samples": 80,
            "subsample": 0.8,
            "colsample_bytree": 1,
            "reg_alpha": 0,
            "reg_lambda": 0,
        }
    },
]



from typing import Optional

import numpy as np
import pandas as pd


def make_submission_df(test_prediction: np.ndarray, sample_submission: Optional[pd.DataFrame] = None,
                       y: Optional[pd.Series] = None) -> pd.DataFrame:
    """
    Make a dataframe formatted as a kaggle competition style.

    Args:
        test_prediction:
            A test prediction to be formatted.
        sample_submission:
            A sample dataframe alined with test data (Usually in Kaggle, it is available as sample_submission.csv).
            The submission file will be created with the same schema as this dataframe.
        y:
            Target variables which is used for inferring the column name. Ignored if ``sample_submission`` is passed.
    Returns:
        The formatted dataframe
    """
    if sample_submission is not None:
        submit_df = sample_submission.copy()

        if test_prediction.ndim > 1 and test_prediction.shape[1] > 1:
            n_id_cols = submit_df.shape[1] - test_prediction.shape[1]
            for i in range(test_prediction.shape[1]):
                submit_df.iloc[:, n_id_cols + i] = test_prediction[:, i]
        else:
            submit_df.iloc[:, -1] = test_prediction
    else:
        submit_df = pd.DataFrame()
        id_col_name = y.index.name if y is not None and y.index.name else 'id'

        submit_df[id_col_name] = np.arange(len(test_prediction))

        if test_prediction.ndim > 1 and test_prediction.shape[1] > 1:
            tgt_col_names = sorted(y.unique()) if y is not None else [str(i) for i in range(test_prediction.shape[1])]
            for i, y in enumerate(tgt_col_names):
                submit_df[y] = test_prediction[:, i]
        else:
            tgt_col_name = y.name if y is not None and y.name else 'target'
            submit_df[tgt_col_name] = test_prediction

    return submit_df



# Modified work:
# -----------------------------------------------------------------------------
# Copyright (c) 2020 Kota Yuhara (@wakamezake)
# -----------------------------------------------------------------------------

# Original work of StratifiedGroupKFold:
# https://github.com/Erotemic/baseline-viame-2018/blob/master/fishnet/util/sklearn_helpers.py
# -----------------------------------------------------------------------------
# Copyright 2018 Jon Crall
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------


import numbers
from datetime import datetime, timedelta
from typing import Iterable, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import sklearn.model_selection as model_selection
from sklearn.model_selection import BaseCrossValidator, KFold, StratifiedKFold
from sklearn.utils.multiclass import type_of_target
from sklearn.utils.validation import check_array
from sklearn.model_selection._split import _BaseKFold


def check_cv(cv: Union[int, Iterable, BaseCrossValidator] = 5,
             y: Optional[Union[pd.Series, np.ndarray]] = None,
             stratified: bool = False,
             random_state: int = 0):
    if cv is None:
        cv = 5
    if isinstance(cv, numbers.Integral):
        if stratified and (y is not None) and (type_of_target(y) in ('binary', 'multiclass')):
            return StratifiedKFold(cv, shuffle=True, random_state=random_state)
        else:
            return KFold(cv, shuffle=True, random_state=random_state)

    return model_selection.check_cv(cv, y, classifier=stratified)


class Take(BaseCrossValidator):
    """ Returns the first N folds of the base validator

    This validator wraps the base validator to take first n folds.

    Args:
        n:
            The number of folds.
        base_validator:
            The base validator to be wrapped.
    Example:
        >>> import numpy as np
        >>> import pandas as pd
        >>> from sklearn.model_selection import KFold
        >>> from nyaggle.validation import Take

        >>> # take the first 3 folds out of 5
        >>> folds = Take(3, KFold(5))
        >>> folds.get_n_splits()
        3
    """

    def __init__(self, n: int, base_validator: BaseCrossValidator):
        self.base_validator = base_validator
        self.n = n

    def get_n_splits(self, X=None, y=None, groups=None):
        return self.n

    def split(self, X, y=None, groups=None):
        """
        Generate indices to split data into training and test set.

        Args:
            X:
                Training data.
            y:
                Target.
            groups:
                Group indices.

        Yields:
            The training set and the testing set indices for that split.
        """
        generator = self.base_validator.split(X, y, groups)
        for i in range(min(self.n, self.base_validator.get_n_splits(X, y, groups))):
            yield next(generator)


class Skip(BaseCrossValidator):
    """ Skips the first N folds and returns the remaining folds

    This validator wraps the base validator to skip first n folds.

    Args:
        n:
            The number of folds to be skipped.
        base_validator:
            The base validator to be wrapped.
    Example:
        >>> import numpy as np
        >>> import pandas as pd
        >>> from sklearn.model_selection import KFold
        >>> from nyaggle.validation import Skip

        >>> # take the last 2 folds out of 5
        >>> folds = Skip(3, KFold(5))
        >>> folds.get_n_splits()
        2
    """

    def __init__(self, n: int, base_validator: BaseCrossValidator):
        self.base_validator = base_validator
        self.n = n

    def get_n_splits(self, X=None, y=None, groups=None):
        return max(self.base_validator.get_n_splits(X, y, groups) - self.n, 0)

    def split(self, X, y=None, groups=None):
        generator = self.base_validator.split(X, y, groups)

        for i in range(self.n):
            next(generator)

        for i in range(self.get_n_splits(X, y, groups)):
            yield next(generator)


class Nth(BaseCrossValidator):
    """ Returns N-th fold of the base validator

    This validator wraps the base validator to take n-th (1-origin) fold.

    Args:
        n:
            The number of folds to be taken.
        base_validator:
            The base validator to be wrapped.
    Example:
        >>> import numpy as np
        >>> import pandas as pd
        >>> from sklearn.model_selection import KFold
        >>> from nyaggle.validation import Nth

        >>> # take the 3rd fold
        >>> folds = Nth(3, KFold(5))
        >>> folds.get_n_splits()
        1
    """

    def __init__(self, n: int, base_validator: BaseCrossValidator):
        assert n > 0, "n is 1-origin and should be greater than 0"
        self.base_validator = Take(1, Skip(n - 1, base_validator))
        self.n = n

    def get_n_splits(self, X=None, y=None, groups=None):
        return 1

    def split(self, X, y=None, groups=None):
        generator = self.base_validator.split(X, y, groups)
        yield next(generator)


class TimeSeriesSplit(BaseCrossValidator):
    """ Time Series cross-validator

    Time Series cross-validator which provides train/test indices to split variable interval time series data.
    This class provides low-level API for time series validation strategy.
    This class is compatible with sklearn's ``BaseCrossValidator`` (base class of ``KFold``, ``GroupKFold`` etc).

    Args:
        source:
            The column name or series of timestamp.
        times:
            Splitting window, where times[i][0] and times[i][1] denotes train and test time interval in (i-1)th fold
            respectively. Each time interval should be pair of datetime or str, and the validator generates indices
            of rows where timestamp is in the half-open interval [start, end).
            For example, if ``times[i][0] = ('2018-01-01', '2018-01-03')``, indices for (i-1)th training data
            will be rows where timestamp value meets ``2018-01-01 <= t < 2018-01-03``.

    Example:
        >>> import numpy as np
        >>> import pandas as pd
        >>> from nyaggle.validation import TimeSeriesSplit
        >>> df = pd.DataFrame()
        >>> df['time'] = pd.date_range(start='2018/1/1', periods=5)

        >>> folds = TimeSeriesSplit('time',
        >>>                         [(('2018-01-01', '2018-01-02'), ('2018-01-02', '2018-01-04')),
        >>>                          (('2018-01-02', '2018-01-03'), ('2018-01-04', '2018-01-06'))])

        >>> folds.get_n_splits()
        2

        >>> splits = folds.split(df)

        >>> train_index, test_index = next(splits)
        >>> train_index
        [0]
        >>> test_index
        [1, 2]

        >>> train_index, test_index = next(splits)
        >>> train_index
        [1]
        >>> test_index
        [3, 4]
    """
    datepair = Tuple[Union[datetime, str], Union[datetime, str]]

    def __init__(self, source: Union[pd.Series, str],
                 times: List[Tuple[datepair, datepair]] = None):
        self.source = source
        self.times = []
        if times:
            for t in times:
                self.add_fold(t[0], t[1])

    def _to_datetime(self, time: Union[str, datetime]):
        return time if isinstance(time, datetime) else pd.to_datetime(time)

    def _to_datetime_tuple(self, time: datepair):
        return self._to_datetime(time[0]), self._to_datetime(time[1])

    def add_fold(self, train_interval: datepair, test_interval: datepair):
        """
        Append 1 split to the validator.

        Args:
            train_interval:
                start and end time of training data.
            test_interval:
                start and end time of test data.
        """
        train_interval = self._to_datetime_tuple(train_interval)
        test_interval = self._to_datetime_tuple(test_interval)
        assert train_interval[1], "train_interval[1] should not be None"
        assert test_interval[0], "test_interval[0] should not be None"

        assert (not train_interval[0]) or (
                    train_interval[0] <= train_interval[1]), "train_interval[0] < train_interval[1]"
        assert (not test_interval[1]) or (test_interval[0] <= test_interval[1]), "test_interval[0] < test_interval[1]"

        self.times.append((train_interval, test_interval))

    def get_n_splits(self, X=None, y=None, groups=None):
        return len(self.times)

    def split(self, X, y=None, groups=None):
        """
        Generate indices to split data into training and test set.

        Args:
            X:
                Training data.
            y:
                Ignored.
            groups:
                Ignored.

        Yields:
            The training set and the testing set indices for that split.
        """
        ts = X[self.source] if isinstance(self.source, str) else self.source

        for train_interval, test_interval in self.times:
            train_mask = ts < train_interval[1]
            if train_interval[0]:
                train_mask = (train_interval[0] <= ts) & train_mask

            test_mask = test_interval[0] <= ts
            if test_interval[1]:
                test_mask = test_mask & (ts < test_interval[1])

            yield np.where(train_mask)[0], np.where(test_mask)[0]


class SlidingWindowSplit(TimeSeriesSplit):
    """ Sliding window time series cross-validator

    Time Series cross-validator which provides train/test indices based on the sliding window to split
    variable interval time series data.
    Splitting for each fold will be as follows:

    .. code-block:: none

      Folds  Training data                                      Testing data
      1      ((train_from-(N-1)*stride, train_to-(N-1)*stride), (test_from-(N-1)*stride, test_to-(N-1)*stride))
      ...    ...                                                ...
      N-1    ((train_from-stride,       train_to-stride),       (test_from-stride,       test_to-stride))
      N      ((train_from,              train_to),              (test_from,              test_to))

    This class is compatible with sklearn's ``BaseCrossValidator`` (base class of ``KFold``, ``GroupKFold`` etc).

    Args:
        source:
            The column name or series of timestamp.
        train_from:
            Start datetime for the training data in the base split.
        train_to:
            End datetime for the training data in the base split.
        test_from:
            Start datetime for the testing data in the base split.
        test_to:
            End datetime for the testing data in the base split.
        n_windows:
            The number of windows (or folds) in the validation.
        stride:
            Time delta between folds.
    """

    date_or_str = Union[datetime, str]

    def __init__(self, source: Union[pd.Series, str],
                 train_from: date_or_str,
                 train_to: date_or_str,
                 test_from: date_or_str,
                 test_to: date_or_str,
                 n_windows: int,
                 stride: timedelta):
        super().__init__(source)

        train_from = self._to_datetime(train_from)
        train_to = self._to_datetime(train_to)
        test_from = self._to_datetime(test_from)
        test_to = self._to_datetime(test_to)

        splits = []

        for i in range(n_windows):
            splits.append(((train_from, train_to), (test_from, test_to)))
            train_from -= stride
            train_to -= stride
            test_from -= stride
            test_to -= stride

        for split in reversed(splits):
            self.add_fold(*split)


class StratifiedGroupKFold(_BaseKFold):
    """ Stratified K-Folds cross-validator with grouping

    Provides train/test indices to split data in train/test sets.
    This cross-validation object is a variation of GroupKFold that returns
    stratified folds. The folds are made by preserving the percentage of
    samples for each class.
    Read more in the :ref:`User Guide <cross_validation>`.

    Args:
        n_splits :
            Number of folds. Must be at least 2.

    Example:
        >>> from pprint import pprint
        >>> rng = np.random.RandomState(0)
        >>> groups = [1, 1, 3, 4, 2, 2, 7, 8, 8]
        >>> y      = [1, 1, 1, 1, 2, 2, 2, 3, 3]
        >>> X = np.empty((len(y), 0))
        >>> self = StratifiedGroupKFold(random_state=rng)
        >>> skf_list = list(self.split(X=X, y=y, groups=groups))
        >>> pprint(skf_list)
        [
            (np.array([2, 3, 4, 5, 6]), np.array([0, 1, 7, 8])),
            (np.array([0, 1, 2, 7, 8]), np.array([3, 4, 5, 6])),
            (np.array([0, 1, 3, 4, 5, 6, 7, 8]), np.array([2])),
        ]
    """

    def __init__(self, n_splits: int = 3, shuffle: bool = False,
                 random_state: Optional[Union[int, np.random.RandomState]] = None):
        super().__init__(n_splits, shuffle=shuffle, random_state=random_state)

    def _make_test_folds(self, X, y=None, groups=None):
        """
        Args:
            X (ndarray):  data
            y (ndarray):  labels(default = None)
            groups (None): (default = None)
        """
        n_splits = self.n_splits
        y = np.asarray(y)
        type_of_target_y = type_of_target(y)
        allowed_target_types = {"binary", "multiclass"}
        if type_of_target_y not in allowed_target_types:
            raise ValueError(
                'Supported target types are: {}. Got {!r} instead.'.format(
                    allowed_target_types, type_of_target_y))

        n_samples = y.shape[0]

        unique_y, y_inversed = np.unique(y, return_inverse=True)
        n_classes = max(unique_y) + 1
        group_to_idxs = self._group_items(groups)
        # unique_groups = list(group_to_idxs.keys())
        group_idxs = list(group_to_idxs.values())
        grouped_y = [y.take(idxs) for idxs in group_idxs]
        grouped_y_counts = np.array([
            np.bincount(y_, minlength=n_classes) for y_ in grouped_y])

        target_freq = grouped_y_counts.sum(axis=0)
        target_ratio = target_freq / target_freq.sum()

        # Greedilly choose the split assignment that minimizes the local
        # * squared differences in target from actual frequencies
        # * and best equalizes the number of items per fold
        # Distribute groups with most members first
        split_freq = np.zeros((n_splits, n_classes))
        # split_ratios = split_freq / split_freq.sum(axis=1)
        split_ratios = np.ones(split_freq.shape) / split_freq.shape[1]
        split_diffs = ((split_freq - target_ratio) ** 2).sum(axis=1)
        sortx = np.argsort(grouped_y_counts.sum(axis=1))[::-1]
        grouped_splitx = []
        for count, group_idx in enumerate(sortx):
            group_freq = grouped_y_counts[group_idx]
            cand_freq = split_freq + group_freq
            cand_ratio = cand_freq / cand_freq.sum(axis=1)[:, None]
            cand_diffs = ((cand_ratio - target_ratio) ** 2).sum(axis=1)
            # Compute loss
            losses = []
            other_diffs = np.array([
                sum(split_diffs[x + 1:]) + sum(split_diffs[:x])
                for x in range(n_splits)
            ])
            # penalize unbalanced splits
            ratio_loss = other_diffs + cand_diffs
            # penalize heavy splits
            freq_loss = split_freq.sum(axis=1)
            denom = freq_loss.sum()
            if denom == 0:
                freq_loss = freq_loss * 0
            else:
                freq_loss = freq_loss / denom
            losses = ratio_loss + freq_loss
            # -------
            splitx = np.argmin(losses)
            split_freq[splitx] = cand_freq[splitx]
            split_ratios[splitx] = cand_ratio[splitx]
            split_diffs[splitx] = cand_diffs[splitx]
            grouped_splitx.append(splitx)

        test_folds = np.empty(n_samples, dtype=np.int)
        for group_idx, splitx in zip(sortx, grouped_splitx):
            idxs = group_idxs[group_idx]
            test_folds[idxs] = splitx

        return test_folds

    @staticmethod
    def _group_items(groups):
        from collections import defaultdict
        group_id_to_items = defaultdict(list)
        pairs = zip(groups, range(len(groups)))
        for k, v in pairs:
            group_id_to_items[k].append(v)
        return group_id_to_items

    def _iter_test_masks(self, X, y=None, groups=None):
        test_folds = self._make_test_folds(X, y, groups)
        for i in range(self.n_splits):
            yield test_folds == i

    def split(self, X, y, groups=None):
        """
        Generate indices to split data into training and test set.
        """
        y = check_array(y, ensure_2d=False, dtype=None)
        return super(StratifiedGroupKFold, self).split(X, y, groups)



from nyaggle.validation.cross_validate import cross_validate
from nyaggle.validation.adversarial_validate import adversarial_validate
from nyaggle.validation.split import \
    check_cv, TimeSeriesSplit, SlidingWindowSplit, Take, Nth, Skip, StratifiedGroupKFold



from typing import List, Optional, Iterable, Union

import category_encoders as ce
import numpy as np
import pandas as pd
from category_encoders.utils import convert_input, convert_input_vector
from sklearn.base import BaseEstimator, clone
from sklearn.model_selection import BaseCrossValidator

from nyaggle.feature.base import BaseFeaturizer
from nyaggle.validation.split import check_cv


class KFoldEncoderWrapper(BaseFeaturizer):
    """KFold Wrapper for sklearn like interface

    This class wraps sklearn's TransformerMixIn (object that has fit/transform/fit_transform methods),
    and call it as K-fold manner.

    Args:
        base_transformer:
            Transformer object to be wrapped.
        cv:
            int, cross-validation generator or an iterable which determines the cross-validation splitting strategy.

            - None, to use the default ``KFold(5, random_state=0, shuffle=True)``,
            - integer, to specify the number of folds in a ``(Stratified)KFold``,
            - CV splitter (the instance of ``BaseCrossValidator``),
            - An iterable yielding (train, test) splits as arrays of indices.
        groups:
            Group labels for the samples. Only used in conjunction with a “Group” cv instance (e.g., ``GroupKFold``).
        return_same_type:
            If True, `transform` and `fit_transform` return the same type as X.
            If False, these APIs always return a numpy array, similar to sklearn's API.
    """

    def __init__(self, base_transformer: BaseEstimator,
                 cv: Optional[Union[int, Iterable, BaseCrossValidator]] = None, return_same_type: bool = True,
                 groups: Optional[pd.Series] = None):
        self.cv = cv
        self.base_transformer = base_transformer

        self.n_splits = None
        self.transformers = None
        self.return_same_type = return_same_type
        self.groups = groups

    def _pre_train(self, y):
        self.cv = check_cv(self.cv, y)
        self.n_splits = self.cv.get_n_splits()
        self.transformers = [clone(self.base_transformer) for _ in range(self.n_splits + 1)]

    def _fit_train(self, X: pd.DataFrame, y: Optional[pd.Series], **fit_params) -> pd.DataFrame:
        if y is None:
            X_ = self.transformers[-1].transform(X)
            return self._post_transform(X_)

        X_ = X.copy()

        for i, (train_index, test_index) in enumerate(self.cv.split(X_, y, self.groups)):
            self.transformers[i].fit(X.iloc[train_index], y.iloc[train_index], **fit_params)
            X_.iloc[test_index, :] = self.transformers[i].transform(X.iloc[test_index])
        self.transformers[-1].fit(X, y, **fit_params)

        return X_

    def _post_fit(self, X: pd.DataFrame, y: pd.Series) -> pd.DataFrame:
        return X

    def _post_transform(self, X: pd.DataFrame) -> pd.DataFrame:
        return X

    def fit(self, X: pd.DataFrame, y: pd.Series):
        """
        Fit models for each fold.

        Args:
            X:
                Data
            y:
                Target
        Returns:
            returns the transformer object.
        """
        self._post_fit(self.fit_transform(X, y), y)
        return self

    def transform(self, X: Union[pd.DataFrame, np.ndarray]) -> Union[pd.DataFrame, np.ndarray]:
        """
        Transform X

        Args:
            X: Data

        Returns:
            Transformed version of X. It will be pd.DataFrame If X is `pd.DataFrame` and return_same_type is True.
        """
        is_pandas = isinstance(X, pd.DataFrame)
        X_ = self._fit_train(X, None)
        X_ = self._post_transform(X_)
        return X_ if self.return_same_type and is_pandas else X_.values

    def fit_transform(self, X: Union[pd.DataFrame, np.ndarray], y: pd.Series = None, **fit_params) \
            -> Union[pd.DataFrame, np.ndarray]:
        """
        Fit models for each fold, then transform X

        Args:
            X:
                Data
            y:
                Target
            fit_params:
                Additional parameters passed to models

        Returns:
            Transformed version of X. It will be pd.DataFrame If X is `pd.DataFrame` and return_same_type is True.
        """
        assert len(X) == len(y)
        self._pre_train(y)

        is_pandas = isinstance(X, pd.DataFrame)
        X = convert_input(X)
        y = convert_input_vector(y, X.index)

        if y.isnull().sum() > 0:
            # y == null is regarded as test data
            X_ = X.copy()
            X_.loc[~y.isnull(), :] = self._fit_train(X[~y.isnull()], y[~y.isnull()], **fit_params)
            X_.loc[y.isnull(), :] = self._fit_train(X[y.isnull()], None, **fit_params)
        else:
            X_ = self._fit_train(X, y, **fit_params)

        X_ = self._post_transform(self._post_fit(X_, y))

        return X_ if self.return_same_type and is_pandas else X_.values


class TargetEncoder(KFoldEncoderWrapper):
    """Target Encoder

    KFold version of category_encoders.TargetEncoder in
    https://contrib.scikit-learn.org/categorical-encoding/targetencoder.html.

    Args:
        cv:
            int, cross-validation generator or an iterable which determines the cross-validation splitting strategy.

            - None, to use the default ``KFold(5, random_state=0, shuffle=True)``,
            - integer, to specify the number of folds in a ``(Stratified)KFold``,
            - CV splitter (the instance of ``BaseCrossValidator``),
            - An iterable yielding (train, test) splits as arrays of indices.
        groups:
            Group labels for the samples. Only used in conjunction with a “Group” cv instance (e.g., ``GroupKFold``).
        cols:
            A list of columns to encode, if None, all string columns will be encoded.
        drop_invariant:
            Boolean for whether or not to drop columns with 0 variance.
        handle_missing:
            Options are ‘error’, ‘return_nan’ and ‘value’, defaults to ‘value’, which returns the target mean.
        handle_unknown:
            Options are ‘error’, ‘return_nan’ and ‘value’, defaults to ‘value’, which returns the target mean.
        min_samples_leaf:
            Minimum samples to take category average into account.
        smoothing:
            Smoothing effect to balance categorical average vs prior. Higher value means stronger regularization.
            The value must be strictly bigger than 0.
        return_same_type:
            If True, ``transform`` and ``fit_transform`` return the same type as X.
            If False, these APIs always return a numpy array, similar to sklearn's API.
    """

    def __init__(self, cv: Optional[Union[Iterable, BaseCrossValidator]] = None,
                 groups: Optional[pd.Series] = None,
                 cols: List[str] = None,
                 drop_invariant: bool = False, handle_missing: str = 'value', handle_unknown: str = 'value',
                 min_samples_leaf: int = 1, smoothing: float = 1.0, return_same_type: bool = True):
        e = ce.TargetEncoder(cols=cols, drop_invariant=drop_invariant, return_df=True,
                             handle_missing=handle_missing,
                             handle_unknown=handle_unknown,
                             min_samples_leaf=min_samples_leaf, smoothing=smoothing)

        super().__init__(e, cv, return_same_type, groups)

    def _post_transform(self, X: pd.DataFrame) -> pd.DataFrame:
        cols = self.transformers[0].cols
        for c in cols:
            X[c] = X[c].astype(float)
        return X



import functools
import os
import warnings
from typing import List, Optional, Union

import pandas as pd
import pyarrow
from tqdm import tqdm


def validate_train_test_difference(train: pd.Series, test: pd.Series):
    # % of nulls
    if test.isnull().mean() == 1.0:
        raise RuntimeError('Error in feature {}: all values in test data is null'.format(train.name))


def validate_feature(df: pd.DataFrame, y: pd.Series):
    if len(y) < len(df):
        # assuming that the first part of the dataframe is train part
        train = df.iloc[:len(y), :]
        test = df.iloc[len(y):, :]
    else:
        train = df[~y.isnull()]
        test = df[y.isnull()]

    for c in df.columns:
        validate_train_test_difference(train[c], test[c])


def save_feature(df: pd.DataFrame, feature_name: Union[int, str], directory: str = './features/',
                 with_csv_dump: bool = False, create_directory: bool = True,
                 reference_target_variable: Optional[pd.Series] = None, overwrite: bool = False):
    """
    Save pandas dataframe as feather-format

    Args:
        df:
            The dataframe to be saved.
        feature_name:
            The name of the feature. The output file will be ``{feature_name}.f``.
        directory:
            The directory where the feature will be stored.
        with_csv_dump:
            If True, the first 1000 lines are dumped to csv file for debug.
        create_directory:
            If True, create directory if not exists.
        reference_target_variable:
            If not None, instant validation will be made on the feature.
        overwrite:
            If False and file already exists, RuntimeError will be raised.
    """
    if create_directory:
        os.makedirs(directory, exist_ok=True)

    if reference_target_variable is not None:
        validate_feature(df, reference_target_variable)

    path = os.path.join(directory, str(feature_name) + '.f')

    if not overwrite and os.path.exists(path):
        raise RuntimeError('File already exists')

    df.to_feather(path)

    if with_csv_dump:
        df.head(1000).to_csv(os.path.join(directory, str(feature_name) + '.csv'), index=False)


def load_feature(feature_name: Union[int, str], directory: str = './features/',
                 ignore_columns: List[str] = None) -> pd.DataFrame:
    """
    Load feature as pandas DataFrame.

    Args:
        feature_name:
            The name of the feature (used in ``save_feature``).
        directory:
            The directory where the feature is stored.
        ignore_columns:
            The list of columns that will be dropped from the loaded dataframe.
    Returns:
        The feature dataframe
    """
    path = os.path.join(directory, str(feature_name) + '.f')

    df = pd.read_feather(path)
    if ignore_columns:
        return df.drop([c for c in ignore_columns if c in df.columns], axis=1)
    else:
        return df


def load_features(base_df: Optional[pd.DataFrame],
                  feature_names: List[Union[int, str]], directory: str = './features/',
                  ignore_columns: List[str] = None, create_directory: bool = True,
                  rename_duplicate: bool = True) -> pd.DataFrame:
    """
    Load features and returns concatenated dataframe

    Args:
        base_df:
            The base dataframe. If not None, resulting dataframe will consist of base and loaded feature columns.
        feature_names:
            The list of feature names to be loaded.
        directory:
            The directory where the feature is stored.
        ignore_columns:
            The list of columns that will be dropped from the loaded dataframe.
        create_directory:
            If True, create directory if not exists.
        rename_duplicate:
            If True, duplicated column name will be renamed automatically (feature name will be used as suffix).
            If False, duplicated columns will be as-is.
    Returns:
        The merged dataframe
    """
    if create_directory:
        os.makedirs(directory, exist_ok=True)

    dfs = [load_feature(f, directory=directory, ignore_columns=ignore_columns) for f in tqdm(feature_names)]

    if base_df is None:
        base_df = dfs[0]
        dfs = dfs[1:]
        feature_names = feature_names[1:]

    columns = list(base_df.columns)

    for df, feature_name in zip(dfs, feature_names):
        if len(df) != len(base_df):
            raise RuntimeError('DataFrame length are different. feature={}'.format(feature_name))

        for c in df.columns:
            if c in columns:
                warnings.warn('A feature name {} is duplicated.'.format(c))

                if rename_duplicate:
                    while c in columns:
                        c += '_' + str(feature_name)
                    warnings.warn('The duplicated name in feature={} will be renamed to {}'.format(feature_name, c))
            columns.append(c)

    concatenated = pd.concat([base_df] + dfs, axis=1)
    concatenated.columns = columns
    return concatenated


def cached_feature(feature_name: Union[int, str], directory: str = './features/', ignore_columns: List[str] = None):
    """
    Decorator to wrap a function which returns pd.DataFrame with a memorizing callable that saves dataframe using
    ``feature_store.save_feature``.

    Args:
        feature_name:
            The name of the feature (used in ``save_feature``).
        directory:
            The directory where the feature is stored.
        ignore_columns:
            The list of columns that will be dropped from the loaded dataframe.

    Example:
        >>> from nyaggle.feature_store import cached_feature
        >>>
        >>> @cached_feature('x')
        >>> def make_feature_x(param) -> pd.DataFrame:
        >>>     print('called')
        >>>     ...
        >>>     return df
        >>>
        >>> x = make_feature_x(...)  # if x.f does not exist, call the function and save result to x.f
        "called"
        >>> x = make_feature_x(...)  # load from file in the second time
    """

    def _decorator(fun):
        @functools.wraps(fun)
        def _decorated_fun(*args, **kwargs):
            try:
                return load_feature(feature_name, directory, ignore_columns)
            except (pyarrow.ArrowIOError, IOError):
                df = fun(*args, **kwargs)
                assert isinstance(df, pd.DataFrame), "returning value of @cached_feature should be pd.DataFrame"
                save_feature(df, feature_name, directory)
                return df

        return _decorated_fun

    return _decorator



from more_itertools import first_true
from typing import Dict, List, Union

from nyaggle.hyper_parameters.catboost import parameters as params_cat
from nyaggle.hyper_parameters.lightgbm import parameters as params_lgb
from nyaggle.hyper_parameters.xgboost import parameters as params_xgb


def _get_hyperparam_byname(param_table: List[Dict], name: str, with_metadata: bool):
    found = first_true(param_table, pred=lambda x: x['name'] == name)
    if found is None:
        raise RuntimeError('Hyperparameter {} not found.'.format(name))

    if with_metadata:
        return found
    else:
        return found['parameters']


def _return(parameter: Union[List[Dict], Dict], with_metadata: bool) -> Union[List[Dict], Dict]:
    if with_metadata:
        return parameter

    if isinstance(parameter, list):
        return [p['parameters'] for p in parameter]
    else:
        return parameter['parameters']


def _get_table(gbdt_type: str = 'lgbm'):
    if gbdt_type == 'lgbm':
        return params_lgb
    elif gbdt_type == 'cat':
        return params_cat
    elif gbdt_type == 'xgb':
        return params_xgb
    raise ValueError('gbdt type should be one of (lgbm, cat, xgb)')


def list_hyperparams(gbdt_type: str = 'lgbm', with_metadata: bool = False) -> List[Dict]:
    """
    List all hyperparameters

    Args:
        gbdt_type:
            The type of gbdt library. ``lgbm``, ``cat``, ``xgb`` can be used.
        with_metadata:
            When set to True, parameters are wrapped by metadata dictionary which contains information about
            source URL, competition name etc.
    Returns:
        A list of hyper-parameters used in Kaggle gold medal solutions
    """
    return _return(_get_table(gbdt_type), with_metadata)


def get_hyperparam_byname(name: str, gbdt_type: str = 'lgbm', with_metadata: bool = False) -> Dict:
    """
    Get a hyperparameter by parameter name

    Args:
        name:
            The name of parameter (e.g. "ieee-2019-10th").
        gbdt_type:
            The type of gbdt library. ``lgbm``, ``cat``, ``xgb`` can be used.
        with_metadata:
            When set to True, parameters are wrapped by metadata dictionary which contains information about
            source URL, competition name etc.
    Returns:
        A hyperparameter dictionary.
    """
    param_table = _get_table(gbdt_type)
    found = first_true(param_table, pred=lambda x: x['name'] == name)
    if found is None:
        raise RuntimeError('Hyperparameter {} not found.'.format(name))

    return _return(found, with_metadata)



parameters = [
    {
        "name": "ieee-2019-17th",
        "url": "https://nbviewer.jupyter.org/github/tmheo/IEEE-Fraud-Detection-17th-Place-Solution/blob/master/notebook/IEEE-17th-Place-Solution-CatBoost-Ensemble.ipynb",
        "competition": "ieee-fraud-detection",
        "rank": 17,
        "metric": "auc",
        "parameters": {
            'learning_rate': 0.07,
            'eval_metric': 'AUC',
            'loss_function': 'Logloss',
            'metric_period': 500,
            'od_wait': 500,
            'depth': 8,
        }
    },
    {
        "name": "elo-2018-11th",
        "url": "https://github.com/kangzhang0709/2019-kaggle-elo-top-11-solution",
        "competition": "elo-merchant-category-recommendation",
        "rank": 11,
        "metric": "rmse",
        "parameters": {
            'learning_rate': 0.01,
            'max_depth': 8,
            'bagging_temperature': 0.8,
            'l2_leaf_reg': 45,
            'od_type': 'Iter'
        }
    },
    {
        "name": "plasticc-2018-3rd",
        "url": "https://github.com/takashioya/plasticc/blob/master/scripts/train.py",
        "competition": "PLAsTiCC-2018",
        "rank": 3,
        "metric": "multi-class log-loss",
        "parameters": {
            'learning_rate': 0.1,
            'depth': 3,
            'loss_function': 'MultiClass',
            'colsample_bylevel': 0.7,
        }
    },
]




import os
import pickle
import time
from collections import namedtuple
from datetime import datetime
from typing import Any, Callable, Dict, Iterable, List, Optional, Type, Union

import pandas as pd
import sklearn.utils.multiclass as multiclass
from sklearn.base import BaseEstimator
from sklearn.metrics import roc_auc_score, mean_squared_error, log_loss
from sklearn.model_selection import BaseCrossValidator

from nyaggle.environment import requires_catboost, requires_lightgbm, requires_xgboost
from nyaggle.experiment.auto_prep import autoprep_gbdt
from nyaggle.experiment.experiment import Experiment
from nyaggle.experiment.hyperparameter_tuner import find_best_lgbm_parameter
from nyaggle.feature_store import load_features
from nyaggle.util import plot_importance, is_gbdt_instance, make_submission_df
from nyaggle.validation.cross_validate import cross_validate
from nyaggle.validation.split import check_cv

ExperimentResult = namedtuple('ExperimentResult',
                              [
                                  'oof_prediction',
                                  'test_prediction',
                                  'metrics',
                                  'models',
                                  'importance',
                                  'time',
                                  'submission_df'
                              ])


class ExpeimentProxy(object):
    __slots__ = ["_obj", "__weakref__"]

    def __init__(self, obj):
        object.__setattr__(self, "_obj", obj)

    def __getattribute__(self, name):
        return getattr(object.__getattribute__(self, "_obj"), name)

    def __setattr__(self, name, value):
        setattr(object.__getattribute__(self, "_obj"), name, value)

    def __enter__(self):
        return self

    def __exit__(self, ex_type, ex_value, trace):
        pass


def run_experiment(model_params: Dict[str, Any],
                   X_train: pd.DataFrame, y: pd.Series,
                   X_test: Optional[pd.DataFrame] = None,
                   logging_directory: str = 'output/{time}',
                   if_exists: str = 'error',
                   eval_func: Optional[Callable] = None,
                   algorithm_type: Union[str, Type[BaseEstimator]] = 'lgbm',
                   fit_params: Optional[Union[Dict[str, Any], Callable]] = None,
                   cv: Optional[Union[int, Iterable, BaseCrossValidator]] = None,
                   groups: Optional[pd.Series] = None,
                   categorical_feature: Optional[List[str]] = None,
                   sample_submission: Optional[pd.DataFrame] = None,
                   submission_filename: Optional[str] = None,
                   type_of_target: str = 'auto',
                   feature_list: Optional[List[Union[int, str]]] = None,
                   feature_directory: Optional[str] = None,
                   inherit_experiment: Optional[Experiment] = None,
                   with_auto_hpo: bool = False,
                   with_auto_prep: bool = False,
                   with_mlflow: bool = False
                   ):
    """
    Evaluate metrics by cross-validation and stores result
    (log, oof prediction, test prediction, feature importance plot and submission file)
    under the directory specified.

    One of the following estimators are used (automatically dispatched by ``type_of_target(y)`` and ``gbdt_type``).

    * LGBMClassifier
    * LGBMRegressor
    * CatBoostClassifier
    * CatBoostRegressor

    The output files are laid out as follows:

    .. code-block:: none

      <logging_directory>/
          log.txt                  <== Logging file
          importance.png           <== Feature importance plot generated by nyaggle.util.plot_importance
          oof_prediction.npy       <== Out of fold prediction in numpy array format
          test_prediction.npy      <== Test prediction in numpy array format
          submission.csv           <== Submission csv file
          metrics.json             <== Metrics
          params.json              <== Parameters
          models/
              fold1                <== The trained model in fold 1
              ...

    Args:
        model_params:
            Parameters passed to the constructor of the classifier/regressor object (i.e. LGBMRegressor).
        X_train:
            Training data. Categorical feature should be casted to pandas categorical type or encoded to integer.
        y:
            Target
        X_test:
            Test data (Optional). If specified, prediction on the test data is performed using ensemble of models.
        logging_directory:
            Path to directory where output of experiment is stored.
        if_exists:
            How to behave if the logging directory already exists.

            - error: Raise a ValueError.
            - replace: Delete logging directory before logging.
            - append: Append to exisitng experiment.
            - rename: Rename current directory by adding "_1", "_2"... prefix
        fit_params:
            Parameters passed to the fit method of the estimator. If dict is passed, the same parameter except
            eval_set passed for each fold. If callable is passed,
            returning value of ``fit_params(fold_id, train_index, test_index)`` will be used for each fold.
        eval_func:
            Function used for logging and calculation of returning scores.
            This parameter isn't passed to GBDT, so you should set objective and eval_metric separately if needed.
            If ``eval_func`` is None, ``roc_auc_score`` or ``mean_squared_error`` is used by default.
        gbdt_type:
            Type of gradient boosting library used. "lgbm" (lightgbm) or "cat" (catboost)
        cv:
            int, cross-validation generator or an iterable which determines the cross-validation splitting strategy.

            - None, to use the default ``KFold(5, random_state=0, shuffle=True)``,
            - integer, to specify the number of folds in a ``(Stratified)KFold``,
            - CV splitter (the instance of ``BaseCrossValidator``),
            - An iterable yielding (train, test) splits as arrays of indices.
        groups:
            Group labels for the samples. Only used in conjunction with a “Group” cv instance (e.g., ``GroupKFold``).
        sample_submission:
            A sample dataframe alined with test data (Usually in Kaggle, it is available as sample_submission.csv).
            The submission file will be created with the same schema as this dataframe.
        submission_filename:
            The name of submission file will be created under logging directory. If ``None``, the basename of the logging
            directory will be used as a filename.
        categorical_feature:
            List of categorical column names. If ``None``, categorical columns are automatically determined by dtype.
        type_of_target:
            The type of target variable. If ``auto``, type is inferred by ``sklearn.utils.multiclass.type_of_target``.
            Otherwise, ``binary``, ``continuous``, or ``multiclass`` are supported.
        feature_list:
            The list of feature ids saved through nyaggle.feature_store module.
        feature_directory:
            The location of features stored. Only used if feature_list is not empty.
        inherit_experiment:
            An experiment object which is used to log results. if not ``None``, all logs in this function are treated
            as a part of this experiment.
        with_auto_prep:
            If True, the input datasets will be copied and automatic preprocessing will be performed on them.
            For example, if ``gbdt_type = 'cat'``, all missing values in categorical features will be filled.
        with_auto_hpo:
            If True, model parameters will be automatically updated using optuna (only available in lightgbm).
        with_mlflow:
            If True, `mlflow tracking <https://www.mlflow.org/docs/latest/tracking.html>`_ is used.
            One instance of ``nyaggle.experiment.Experiment`` corresponds to one run in mlflow.
            Note that all output
            mlflow's directory (``mlruns`` by default).
    :return:
        Namedtuple with following members

        * oof_prediction:
            numpy array, shape (len(X_train),) Predicted value on Out-of-Fold validation data.
        * test_prediction:
            numpy array, shape (len(X_test),) Predicted value on test data. ``None`` if X_test is ``None``
        * metrics:
            list of float, shape(nfolds+1) ``metrics[i]`` denotes validation score in i-th fold.
            ``metrics[-1]`` is overall score.
        * models:
            list of objects, shape(nfolds) Trained models for each folds.
        * importance:
            list of pd.DataFrame, feature importance for each fold (type="gain").
        * time:
            Training time in seconds.
        * submit_df:
            The dataframe saved as submission.csv
    """
    start_time = time.time()
    cv = check_cv(cv, y)

    if feature_list:
        X = pd.concat([X_train, X_test]) if X_test is not None else X_train
        X.reset_index(drop=True, inplace=True)
        X = load_features(X, feature_list, directory=feature_directory)
        ntrain = len(X_train)
        X_train, X_test = X.iloc[:ntrain, :], X.iloc[ntrain:, :].reset_index(drop=True)

    _check_input(X_train, y, X_test)

    if categorical_feature is None:
        categorical_feature = [c for c in X_train.columns if X_train[c].dtype.name in ['object', 'category']]

    if type_of_target == 'auto':
        type_of_target = multiclass.type_of_target(y)
    model_type, eval_func, cat_param_name = _dispatch_models(algorithm_type, type_of_target, eval_func)

    if with_auto_prep:
        assert algorithm_type in ('cat', 'xgb', 'lgbm'), "with_auto_prep is only supported for gbdt"
        X_train, X_test = autoprep_gbdt(algorithm_type, X_train, X_test, categorical_feature)

    logging_directory = logging_directory.format(time=datetime.now().strftime('%Y%m%d_%H%M%S'))

    if inherit_experiment is not None:
        experiment = ExpeimentProxy(inherit_experiment)
    else:
        experiment = Experiment(logging_directory, if_exists=if_exists, with_mlflow=with_mlflow)

    with experiment as exp:
        exp.log('Algorithm: {}'.format(algorithm_type))
        exp.log('Experiment: {}'.format(exp.logging_directory))
        exp.log('Params: {}'.format(model_params))
        exp.log('Features: {}'.format(list(X_train.columns)))
        exp.log_param('algorithm_type', algorithm_type)
        exp.log_param('num_features', X_train.shape[1])
        if callable(fit_params):
            exp.log_param('fit_params', str(fit_params))
        else:
            exp.log_dict('fit_params', fit_params)
        exp.log_dict('model_params', model_params)
        if feature_list is not None:
            exp.log_param('features', feature_list)

        if with_auto_hpo:
            assert algorithm_type == 'lgbm', 'auto-tuning is only supported for LightGBM'
            model_params = find_best_lgbm_parameter(model_params, X_train, y, cv=cv, groups=groups,
                                                    type_of_target=type_of_target)
            exp.log_param('model_params_tuned', model_params)

        exp.log('Categorical: {}'.format(categorical_feature))

        models = [model_type(**model_params) for _ in range(cv.get_n_splits())]

        if fit_params is None:
            fit_params = {}
        if cat_param_name is not None and not callable(fit_params) and cat_param_name not in fit_params:
            fit_params[cat_param_name] = categorical_feature

        if isinstance(fit_params, Dict):
            exp.log_params(fit_params)

        result = cross_validate(models, X_train=X_train, y=y, X_test=X_test, cv=cv, groups=groups,
                                logger=exp.get_logger(), eval_func=eval_func, fit_params=fit_params,
                                type_of_target=type_of_target)

        # save oof
        exp.log_numpy('oof_prediction', result.oof_prediction)
        exp.log_numpy('test_prediction', result.test_prediction)

        for i in range(cv.get_n_splits()):
            exp.log_metric('Fold {}'.format(i + 1), result.scores[i])
        exp.log_metric('Overall', result.scores[-1])

        # save importance plot
        if result.importance:
            importance = pd.concat(result.importance)
            plot_file_path = os.path.join(exp.logging_directory, 'importance.png')
            plot_importance(importance, plot_file_path)
            exp.log_artifact(plot_file_path)

        # save trained model
        for i, model in enumerate(models):
            _save_model(model, exp.logging_directory, i + 1, exp)

        # save submission.csv
        submit_df = None
        if X_test is not None:
            submit_df = make_submission_df(result.test_prediction, sample_submission, y)
            exp.log_dataframe(submission_filename or os.path.basename(exp.logging_directory), submit_df, 'csv')

        elapsed_time = time.time() - start_time

        return ExperimentResult(result.oof_prediction, result.test_prediction,
                                result.scores, models, result.importance, elapsed_time, submit_df)


def _dispatch_eval_func(target_type: str, custom_eval: Optional[Callable] = None):
    default_eval_func = {
        'binary': roc_auc_score,
        'multiclass': log_loss,
        'continuous': mean_squared_error
    }
    return custom_eval if custom_eval is not None else default_eval_func[target_type]


def _dispatch_gbdt_class(algorithm_type: str, type_of_target: str):
    is_regression = type_of_target == 'continuous'

    if algorithm_type == 'lgbm':
        requires_lightgbm()
        from lightgbm import LGBMClassifier, LGBMRegressor
        return LGBMRegressor if is_regression else LGBMClassifier
    elif algorithm_type == 'cat':
        requires_catboost()
        from catboost import CatBoostClassifier, CatBoostRegressor
        return CatBoostRegressor if is_regression else CatBoostClassifier
    else:
        requires_xgboost()
        assert algorithm_type == 'xgb'
        from xgboost import XGBClassifier, XGBRegressor
        return XGBRegressor if is_regression else XGBClassifier


def _dispatch_models(algorithm_type: Union[str, Type[BaseEstimator]],
                     target_type: str, custom_eval: Optional[Callable] = None):
    if not isinstance(algorithm_type, str):
        assert issubclass(algorithm_type, BaseEstimator), "algorithm_type should be str or subclass of BaseEstimator"
        return algorithm_type, _dispatch_eval_func(target_type, custom_eval), None

    cat_features = {
        'lgbm': 'categorical_feature',
        'cat': 'cat_features',
        'xgb': None
    }

    gbdt_class = _dispatch_gbdt_class(algorithm_type, target_type)
    eval_func = _dispatch_eval_func(target_type, custom_eval)

    return gbdt_class, eval_func, cat_features[algorithm_type]


def _save_model(model: BaseEstimator, logging_directory: str, fold: int, exp: Experiment):
    model_dir = os.path.join(logging_directory, 'models')
    os.makedirs(model_dir, exist_ok=True)
    path = os.path.join(model_dir, 'fold{}'.format(fold))

    if is_gbdt_instance(model, 'lgbm'):
        model.booster_.save_model(path)
    elif is_gbdt_instance(model, ('xgb', 'cat')):
        model.save_model(path)
    else:
        with open(path, "wb") as f:
            pickle.dump(model, f)

    exp.log_artifact(path)


def _check_input(X_train: pd.DataFrame, y: pd.Series,
                 X_test: Optional[pd.DataFrame] = None):
    assert len(X_train) == len(y), "length of X_train and y are different. len(X_train) = {}, len(y) = {}".format(
        len(X_train), len(y)
    )

    if X_test is not None:
        assert list(X_train.columns) == list(X_test.columns), "columns are different between X_train and X_test"



import copy
import warnings
from typing import Dict, Iterable, Optional, Union

import optuna.integration.lightgbm as optuna_lgb
import pandas as pd
import sklearn.utils.multiclass as multiclass
from sklearn.model_selection import BaseCrossValidator

from nyaggle.validation.split import check_cv


def find_best_lgbm_parameter(base_param: Dict, X: pd.DataFrame, y: pd.Series,
                             cv: Optional[Union[int, Iterable, BaseCrossValidator]] = None,
                             groups: Optional[pd.Series] = None,
                             time_budget: Optional[int] = None,
                             type_of_target: str = 'auto') -> Dict:
    """
    Search hyperparameter for lightgbm using optuna.

    Args:
        base_param:
            Base parameters passed to lgb.train.
        X:
            Training data.
        y:
            Target
        cv:
            int, cross-validation generator or an iterable which determines the cross-validation splitting strategy.
        groups:
            Group labels for the samples. Only used in conjunction with a “Group” cv instance (e.g., ``GroupKFold``).
        time_budget:
            Time budget for tuning (in seconds).
        type_of_target:
            The type of target variable. If ``auto``, type is inferred by ``sklearn.utils.multiclass.type_of_target``.
            Otherwise, ``binary``, ``continuous``, or ``multiclass`` are supported.

    Returns:
        The best parameters found
    """
    cv = check_cv(cv, y)

    if type_of_target == 'auto':
        type_of_target = multiclass.type_of_target(y)

    train_index, test_index = next(cv.split(X, y, groups))

    dtrain = optuna_lgb.Dataset(X.iloc[train_index], y.iloc[train_index])
    dvalid = optuna_lgb.Dataset(X.iloc[test_index], y.iloc[test_index])

    params = copy.deepcopy(base_param)
    if 'early_stopping_rounds' not in params:
        params['early_stopping_rounds'] = 100

    if params.get('feature_pre_filter'):
        warnings.warn("feature_pre_filter will be set to False to tune min_data_in_leaf.")
    params['feature_pre_filter'] = False

    if not any([p in params for p in ('num_iterations', 'num_iteration',
                                      'num_trees', 'num_tree',
                                      'num_rounds', 'num_round')]):
        params['num_iterations'] = params.get('n_estimators', 10000)

    if 'objective' not in params:
        tot_to_objective = {
            'binary': 'binary',
            'continuous': 'regression',
            'multiclass': 'multiclass'
        }
        params['objective'] = tot_to_objective[type_of_target]

    if 'metric' not in params and 'objective' in params:
        if params['objective'] in ['regression', 'regression_l2', 'l2', 'mean_squared_error', 'mse', 'l2_root',
                                   'root_mean_squared_error', 'rmse']:
            params['metric'] = 'l2'
        if params['objective'] in ['regression_l1', 'l1', 'mean_absolute_error', 'mae']:
            params['metric'] = 'l1'
        if params['objective'] in ['binary']:
            params['metric'] = 'binary_logloss'
        if params['objective'] in ['multiclass']:
            params['metric'] = 'multi_logloss'

    if not any([p in params for p in ('verbose', 'verbosity')]):
        params['verbosity'] = -1

    model = optuna_lgb.train(params, dtrain, valid_sets=[dvalid], verbose_eval=0,
                             time_budget=time_budget)

    return model.params



import json
import numbers
import os
import shutil
import uuid
import warnings
from logging import getLogger, FileHandler, DEBUG, Logger
from typing import Dict, Optional

import numpy as np
import pandas as pd

from nyaggle.environment import requires_mlflow

MLFLOW_KEY_LENGTH_LIMIT = 250
MLFLOW_VALUE_LENGTH_LIMIT = 250


def _sanitize_mlflow_param(param, limit):
    if len(str(param)) > limit:
        warnings.warn('Length of param exceeds limit {}. It will be truncated. value: {}'.format(limit, param))
        param = str(param)[:limit]
    return param


def _check_directory(directory: str, if_exists: str) -> str:
    if os.path.exists(directory):
        if if_exists == 'error':
            raise ValueError('directory {} already exists.'.format(directory))
        elif if_exists == 'replace':
            warnings.warn(
                'directory {} already exists. It will be replaced by the new result'.format(directory))

            existing_run_id = _try_to_get_existing_mlflow_run_id(directory)
            if existing_run_id is not None:
                requires_mlflow()
                import mlflow
                mlflow.delete_run(existing_run_id)

            shutil.rmtree(directory, ignore_errors=True)
        elif if_exists == 'rename':
            postfix_index = 1

            while os.path.exists(directory + '_' + str(postfix_index)):
                postfix_index += 1

            directory += '_' + str(postfix_index)
            warnings.warn('directory is renamed to {} because the original directory already exists.'.format(directory))
    return directory


def _sanitize(v):
    return v if isinstance(v, numbers.Number) else str(v)


def _try_to_get_existing_mlflow_run_id(logging_directory: str) -> Optional[str]:
    mlflow_path = os.path.join(logging_directory, 'mlflow.json')
    if os.path.exists(mlflow_path):
        with open(mlflow_path, 'r') as f:
            mlflow_metadata = json.load(f)
            return mlflow_metadata['run_id']
    return None


class Experiment(object):
    """Minimal experiment logger for Kaggle

    This module provides minimal functionality for tracking experiments.
    The output files are laid out as follows:

    .. code-block:: none

      <logging_directory>/
          log.txt       <== Output of log
          metrics.json  <== Output of log_metric(s), format: name,score
          params.json   <== Output of log_param(s), format: key,value
          mlflow.json   <== mlflow's run_id, experiment_id and artifact_uri (logged if with_mlflow=True)


    You can add numpy array and pandas dataframe under the directory through ``log_numpy`` and ``log_dataframe``.

    Args:
        logging_directory:
            Path to directory where output is stored.
        custom_logger:
            A custom logger to be used instead of default logger.
        with_mlflow:
            If True, `mlflow tracking <https://www.mlflow.org/docs/latest/tracking.html>`_ is used.
            One instance of ``nyaggle.experiment.Experiment`` corresponds to one run in mlflow.
            Note that all output files are located both ``logging_directory`` and
            mlflow's directory (``mlruns`` by default).
        if_exists:
            How to behave if the logging directory already exists.

            - error: Raise a ValueError.
            - replace: Delete logging directory before logging.
            - append: Append to exisitng experiment.
            - rename: Rename current directory by adding "_1", "_2"... prefix
    Example:
        >>> import numpy as np
        >>> import pandas as pd
        >>> from nyaggle.experiment import Experiment
        >>>
        >>> with Experiment(logging_directory='./output/') as exp:
        >>>     # log key-value pair as a parameter
        >>>     exp.log_param('lr', 0.01)
        >>>     exp.log_param('optimizer', 'adam')
        >>>
        >>>     # log text
        >>>     exp.log('blah blah blah')
        >>>
        >>>     # log metric
        >>>     exp.log_metric('CV', 0.85)
        >>>
        >>>     # log dictionary with flattening keys
        >>>     exp.log_dict('params', {'X': 3, 'Y': {'Z': 'foobar'}})
        >>>
        >>>     # log numpy ndarray, pandas dafaframe and any artifacts
        >>>     exp.log_numpy('predicted', np.zeros(1))
        >>>     exp.log_dataframe('submission', pd.DataFrame(), file_format='csv')
        >>>     exp.log_artifact('path-to-your-file')
    """

    def __init__(self,
                 logging_directory: str,
                 custom_logger: Optional[Logger] = None,
                 with_mlflow: bool = False,
                 if_exists: str = 'error'
                 ):
        logging_directory = _check_directory(logging_directory, if_exists)
        os.makedirs(logging_directory, exist_ok=True)

        self.logging_directory = logging_directory
        self.with_mlflow = with_mlflow

        if custom_logger is not None:
            self.logger = custom_logger
            self.is_custom = True
        else:
            self.logger = getLogger(str(uuid.uuid4()))
            self.log_path = os.path.join(logging_directory, 'log.txt')
            self.logger.addHandler(FileHandler(self.log_path))
            self.logger.setLevel(DEBUG)
            self.is_custom = False
        self.metrics = self._load_dict('metrics.json')
        self.params = self._load_dict('params.json')
        self.inherit_existing_run = False

        if self.with_mlflow:
            requires_mlflow()
            self.mlflow_run_id = _try_to_get_existing_mlflow_run_id(logging_directory)
            if self.mlflow_run_id is not None:
                self.mlflow_run_name = None
            else:
                self.mlflow_run_name = logging_directory

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, ex_type, ex_value, trace):
        self.stop()

    @classmethod
    def continue_from(cls, logging_directory: str, with_mlflow: bool = False):
        return cls(logging_directory=logging_directory, if_exists='append', with_mlflow=with_mlflow)

    def start(self):
        """
        Start a new experiment.
        """
        if self.with_mlflow:
            import mlflow

            if mlflow.active_run() is not None:
                active_run = mlflow.active_run()
                self.inherit_existing_run = True
            else:
                active_run = mlflow.start_run(run_name=self.mlflow_run_name, run_id=self.mlflow_run_id)
            mlflow_metadata = {
                'artifact_uri': active_run.info.artifact_uri,
                'experiment_id': active_run.info.experiment_id,
                'run_id': active_run.info.run_id
            }
            self.mlflow_run_id = active_run.info.run_id
            with open(os.path.join(self.logging_directory, 'mlflow.json'), 'w') as f:
                json.dump(mlflow_metadata, f, indent=4)

    def _load_dict(self, filename: str) -> Dict:
        try:
            path = os.path.join(self.logging_directory, filename)
            with open(path, 'r') as f:
                return json.load(f)
        except IOError:
            self.logger.warning('failed to load file: {}'.format(filename))
            return {}

    def _save_dict(self, obj: Dict, filename: str):
        try:
            path = os.path.join(self.logging_directory, filename)
            with open(path, 'w') as f:
                json.dump(obj, f, indent=2)
        except IOError:
            self.logger.warning('failed to save file: {}'.format(filename))

    def stop(self):
        """
        Stop current experiment.
        """
        self._save_dict(self.metrics, 'metrics.json')
        self._save_dict(self.params, 'params.json')

        if not self.is_custom:
            for h in self.logger.handlers:
                h.close()

        if self.with_mlflow:
            import mlflow
            from mlflow.exceptions import MlflowException

            try:
                mlflow.log_artifact(self.log_path)
                mlflow.log_artifact(os.path.join(self.logging_directory, 'metrics.json'))
                mlflow.log_artifact(os.path.join(self.logging_directory, 'params.json'))
            except MlflowException as e:
                warnings.warn('Error in saving artifacts to mlflow. The result may not be saved.: {}'.format(e))
            if not self.inherit_existing_run:
                mlflow.end_run()

    def get_logger(self) -> Logger:
        """
        Get logger used in this experiment.

        Returns:
            logger object
        """
        return self.logger

    def get_run(self):
        """
        Get mlflow's currently active run, or None if ``with_mlflow = False``.

        Returns:
            active Run
        """
        if not self.with_mlflow:
            return None

        import mlflow
        return mlflow.active_run()

    def log(self, text: str):
        """
        Logs a message on the logger for the experiment.

        Args:
            text:
                The message to be written.
        """
        self.logger.info(text)

    def log_param(self, key, value):
        """
        Logs a key-value pair for the experiment.

        Args:
            key: parameter name
            value: parameter value
        """
        key = _sanitize(key)
        value = _sanitize(value)
        self.params[key] = value

        if self.with_mlflow:
            import mlflow
            from mlflow.exceptions import MlflowException

            key_mlflow = _sanitize_mlflow_param(key, MLFLOW_KEY_LENGTH_LIMIT)
            value_mlflow = _sanitize_mlflow_param(value, MLFLOW_VALUE_LENGTH_LIMIT)

            try:
                mlflow.log_param(key_mlflow, value_mlflow)
            except MlflowException as e:
                warnings.warn('Error in logging parameter {} to mlflow. Skipped. {}'.format(key, e))

    def log_params(self, params: Dict):
        """
        Logs a batch of params for the experiments.

        Args:
            params: dictionary of parameters
        """
        for k, v in params.items():
            self.log_param(k, v)

    def log_dict(self, name: str, value: Dict, separator: str = '.'):
        """
        Logs a dictionary as parameter with flatten format.

        Args:
            name: Parameter name
            value: Parameter value
            separator: Separating character used to concatanate keys
        Examples:
            >>> with Experiment('./') as e:
            >>>     e.log_dict('a', {'b': 1, 'c': 'd'})
            >>>     print(e.params)
            { 'a.b': 1, 'a.c': 'd' }
        """

        if value is None:
            self.log_param(name, value)
            return

        def _flatten(d: Dict, prefix: str, separator: str) -> Dict:
            items = []
            for k, v in d.items():
                child_key = prefix + separator + str(k) if prefix else str(k)
                if isinstance(v, Dict) and v:
                    items.extend(_flatten(v, child_key, separator).items())
                else:
                    items.append((child_key, v))
            return dict(items)

        value = _flatten(value, name, separator)
        self.log_params(value)

    def log_metric(self, name: str, score: float):
        """
        Log a metric under the logging directory.

        Args:
            name:
                Metric name.
            score:
                Metric value.
        """
        name = _sanitize(name)
        score = _sanitize(score)
        self.metrics[name] = score

        if self.with_mlflow:
            import mlflow
            from mlflow.exceptions import MlflowException

            try:
                mlflow.log_metric(name, score)
            except MlflowException as e:
                warnings.warn('Error in logging metric {} to mlflow. Skipped. {}'.format(name, e))

    def log_metrics(self, metrics: Dict):
        """
        Log a batch of metrics under the logging directory.

        Args:
            metrics: dictionary of metrics.
        """
        for k, v in metrics.items():
            self.log_metric(k, v)

    def log_numpy(self, name: str, array: np.ndarray):
        """
        Log a numpy ndarray under the logging directory.

        Args:
            name:
                Name of the file. A .npy extension will be appended to the file name if it does not already have one.
            array:
                Array data to be saved.
        """
        path = os.path.join(self.logging_directory, name)
        np.save(path, array)

        if self.with_mlflow:
            import mlflow
            mlflow.log_artifact(path + '.npy')

    def log_dataframe(self, name: str, df: pd.DataFrame, file_format: str = 'feather'):
        """
        Log a pandas dataframe under the logging directory.

        Args:
            name:
                Name of the file. A ``.f`` or ``.csv`` extension will be appended to the file name
                if it does not already have one.
            df:
                A dataframe to be saved.
            file_format:
                A format of output file. ``csv`` and ``feather`` are supported.
        """
        path = os.path.join(self.logging_directory, name)
        if file_format == 'feather':
            if not path.endswith('.f'):
                path += '.f'
            df.to_feather(path)
        elif file_format == 'csv':
            if not path.endswith('.csv'):
                path += '.csv'
            df.to_csv(path, index=False)
        else:
            raise RuntimeError('format not supported')

        if self.with_mlflow:
            import mlflow
            mlflow.log_artifact(path)

    def log_artifact(self, src_file_path: str):
        """
        Make a copy of the file under the logging directory.

        Args:
            src_file_path:
                Path of the file. If path is not a child of the logging directory, the file will be copied.
                If ``with_mlflow`` is True, ``mlflow.log_artifact`` will be called (then another copy will be made).
        """
        logging_path = os.path.abspath(self.logging_directory)
        src_file_path = os.path.abspath(src_file_path)

        if os.path.commonpath([logging_path]) != os.path.commonpath([logging_path, src_file_path]):
            src_file = os.path.basename(src_file_path)
            shutil.copy(src_file, self.logging_directory)

        if self.with_mlflow:
            import mlflow
            mlflow.log_artifact(src_file_path)


def add_leaderboard_score(logging_directory: str, score: float):
    """
    Record leaderboard score to the existing experiment directory.

    Args:
        logging_directory:
            The directory to be added
        score:
            Leaderboard score
    """
    with Experiment.continue_from(logging_directory) as e:
        e.log_metric('LB', score)



import copy
import time
import warnings
from collections import namedtuple
from logging import Logger, getLogger
from typing import Any, Callable, Dict, Iterable, List, Optional, Union

import numpy as np
import pandas as pd
import sklearn.utils.multiclass as multiclass
from category_encoders.utils import convert_input, convert_input_vector
from sklearn.base import BaseEstimator
from sklearn.model_selection import BaseCrossValidator

from nyaggle.util.traits import is_gbdt_instance
from nyaggle.validation.split import check_cv

CVResult = namedtuple('CVResult', ['oof_prediction', 'test_prediction', 'scores', 'importance'])


def cross_validate(estimator: Union[BaseEstimator, List[BaseEstimator]],
                   X_train: Union[pd.DataFrame, np.ndarray], y: Union[pd.Series, np.ndarray],
                   X_test: Union[pd.DataFrame, np.ndarray] = None,
                   cv: Optional[Union[int, Iterable, BaseCrossValidator]] = None,
                   groups: Optional[pd.Series] = None,
                   eval_func: Optional[Callable] = None, logger: Optional[Logger] = None,
                   on_each_fold: Optional[Callable[[int, BaseEstimator, pd.DataFrame, pd.Series], None]] = None,
                   fit_params: Optional[Union[Dict[str, Any], Callable]] = None,
                   importance_type: str = 'gain',
                   early_stopping: bool = True,
                   type_of_target: str = 'auto') -> CVResult:
    """
    Evaluate metrics by cross-validation. It also records out-of-fold prediction and test prediction.

    Args:
        estimator:
            The object to be used in cross-validation. For list inputs, ``estimator[i]`` is trained on i-th fold.
        X_train:
            Training data
        y:
            Target
        X_test:
            Test data (Optional). If specified, prediction on the test data is performed using ensemble of models.
        cv:
            int, cross-validation generator or an iterable which determines the cross-validation splitting strategy.

            - None, to use the default ``KFold(5, random_state=0, shuffle=True)``,
            - integer, to specify the number of folds in a ``(Stratified)KFold``,
            - CV splitter (the instance of ``BaseCrossValidator``),
            - An iterable yielding (train, test) splits as arrays of indices.
        groups:
            Group labels for the samples. Only used in conjunction with a “Group” cv instance (e.g., ``GroupKFold``).
        eval_func:
            Function used for logging and returning scores
        logger:
            logger
        on_each_fold:
            called for each fold with (idx_fold, model, X_fold, y_fold)
        fit_params:
            Parameters passed to the fit method of the estimator
        importance_type:
            The type of feature importance to be used to calculate result.
            Used only in ``LGBMClassifier`` and ``LGBMRegressor``.
        early_stopping:
            If ``True``, ``eval_set`` will be added to ``fit_params`` for each fold.
            ``early_stopping_rounds = 100`` will also be appended to fit_params if it does not already have one.
        type_of_target:
            The type of target variable. If ``auto``, type is inferred by ``sklearn.utils.multiclass.type_of_target``.
            Otherwise, ``binary``, ``continuous``, or ``multiclass`` are supported.
    Returns:
        Namedtuple with following members

        * oof_prediction (numpy array, shape (len(X_train),)):
            The predicted value on put-of-Fold validation data.
        * test_prediction (numpy array, hape (len(X_test),)):
            The predicted value on test data. ``None`` if X_test is ``None``.
        * scores (list of float, shape (nfolds+1,)):
            ``scores[i]`` denotes validation score in i-th fold.
            ``scores[-1]`` is the overall score. `None` if eval is not specified.
        * importance (list of pandas DataFrame, shape (nfolds,)):
            ``importance[i]`` denotes feature importance in i-th fold model.
            If the estimator is not GBDT, empty array is returned.

    Example:
        >>> from sklearn.datasets import make_regression
        >>> from sklearn.linear_model import Ridge
        >>> from sklearn.metrics import mean_squared_error
        >>> from nyaggle.validation import cross_validate

        >>> X, y = make_regression(n_samples=8)
        >>> model = Ridge(alpha=1.0)
        >>> pred_oof, pred_test, scores, _ = \
        >>>     cross_validate(model,
        >>>                    X_train=X[:3, :],
        >>>                    y=y[:3],
        >>>                    X_test=X[3:, :],
        >>>                    cv=3,
        >>>                    eval_func=mean_squared_error)
        >>> print(pred_oof)
        [-101.1123267 ,   26.79300693,   17.72635528]
        >>> print(pred_test)
        [-10.65095894 -12.18909059 -23.09906427 -17.68360714 -20.08218267]
        >>> print(scores)
        [71912.80290003832, 15236.680239881942, 15472.822033121925, 34207.43505768073]
    """
    cv = check_cv(cv, y)
    n_output_cols = 1
    if type_of_target == 'auto':
        type_of_target = multiclass.type_of_target(y)
    if type_of_target == 'multiclass':
        n_output_cols = y.nunique(dropna=True)

    if isinstance(estimator, list):
        assert len(estimator) == cv.get_n_splits(), "Number of estimators should be same to nfolds."

    X_train = convert_input(X_train)
    y = convert_input_vector(y, X_train.index)
    if X_test is not None:
        X_test = convert_input(X_test)

    if not isinstance(estimator, list):
        estimator = [estimator] * cv.get_n_splits()

    assert len(estimator) == cv.get_n_splits()

    if logger is None:
        logger = getLogger(__name__)

    def _predict(model: BaseEstimator, x: pd.DataFrame, _type_of_target: str):
        if _type_of_target in ('binary', 'multiclass'):
            if hasattr(model, "predict_proba"):
                proba = model.predict_proba(x)
            elif hasattr(model, "decision_function"):
                warnings.warn('Since {} does not have predict_proba method, '
                              'decision_function is used for the prediction instead.'.format(type(model)))
                proba = model.decision_function(x)
            else:
                raise RuntimeError('Estimator in classification problem should have '
                                   'either predict_proba or decision_function')
            if proba.ndim == 1:
                return proba
            else:
                return proba[:, 1] if proba.shape[1] == 2 else proba
        else:
            return model.predict(x)

    oof = np.zeros((len(X_train), n_output_cols)) if n_output_cols > 1 else np.zeros(len(X_train))
    evaluated = np.full(len(X_train), False)
    test = None
    if X_test is not None:
        test = np.zeros((len(X_test), n_output_cols)) if n_output_cols > 1 else np.zeros(len(X_test))

    scores = []
    eta_all = []
    importance = []

    for n, (train_idx, valid_idx) in enumerate(cv.split(X_train, y, groups)):
        start_time = time.time()

        train_x, train_y = X_train.iloc[train_idx], y.iloc[train_idx]
        valid_x, valid_y = X_train.iloc[valid_idx], y.iloc[valid_idx]

        if fit_params is None:
            fit_params_fold = {}
        elif callable(fit_params):
            fit_params_fold = fit_params(n, train_idx, valid_idx)
        else:
            fit_params_fold = copy.copy(fit_params)

        if is_gbdt_instance(estimator[n], ('lgbm', 'cat', 'xgb')):
            if early_stopping:
                if 'eval_set' not in fit_params_fold:
                    fit_params_fold['eval_set'] = [(valid_x, valid_y)]
                if 'early_stopping_rounds' not in fit_params_fold:
                    fit_params_fold['early_stopping_rounds'] = 100

            estimator[n].fit(train_x, train_y, **fit_params_fold)
        else:
            estimator[n].fit(train_x, train_y, **fit_params_fold)

        oof[valid_idx] = _predict(estimator[n], valid_x, type_of_target)
        evaluated[valid_idx] = True

        if X_test is not None:
            test += _predict(estimator[n], X_test, type_of_target)

        if on_each_fold is not None:
            on_each_fold(n, estimator[n], train_x, train_y)

        if is_gbdt_instance(estimator[n], ('lgbm', 'cat', 'xgb')):
            importance.append(_get_gbdt_importance(estimator[n], list(X_train.columns), importance_type))

        if eval_func is not None:
            score = eval_func(valid_y, oof[valid_idx])
            scores.append(score)
            logger.info('Fold {} score: {}'.format(n, score))

        elapsed = time.time() - start_time
        eta_all.append(elapsed)
        logger.debug('{:.3f} sec / fold'.format(elapsed))

    if eval_func is not None:
        score = eval_func(y.loc[evaluated], oof[evaluated])
        scores.append(score)
        logger.info('Overall score: {}'.format(score))

    if X_test is not None:
        predicted = test / cv.get_n_splits(X_train, y, groups)
    else:
        predicted = None

    return CVResult(oof, predicted, scores, importance)


def _get_gbdt_importance(gbdt_model: BaseEstimator, features: List[str],
                         importance_type: str) -> pd.DataFrame:
    df = pd.DataFrame()

    df['feature'] = features

    if is_gbdt_instance(gbdt_model, 'cat'):
        df['importance'] = gbdt_model.get_feature_importance()
    elif is_gbdt_instance(gbdt_model, 'xgb'):
        df['importance'] = gbdt_model.feature_importances_
    elif is_gbdt_instance(gbdt_model, 'lgbm'):
        df['importance'] = gbdt_model.booster_.feature_importance(importance_type=importance_type)

    return df



from typing import Callable, Iterable, List, Union, Optional

import numpy as np
import pandas as pd
import sklearn.utils.multiclass as multiclass
from category_encoders.utils import convert_input, convert_input_vector
from sklearn.base import BaseEstimator
from sklearn.linear_model import LogisticRegression, Ridge
from sklearn.model_selection import BaseCrossValidator, GridSearchCV

from nyaggle.ensemble.common import EnsembleResult
from nyaggle.validation import cross_validate


def stacking(test_predictions: List[np.ndarray],
             oof_predictions: List[np.ndarray],
             y: pd.Series,
             estimator: Optional[BaseEstimator] = None,
             cv: Optional[Union[int, Iterable, BaseCrossValidator]] = None,
             groups: Optional[pd.Series] = None,
             type_of_target: str = 'auto',
             eval_func: Optional[Callable] = None) -> EnsembleResult:
    """
    Perform stacking on predictions.

    Args:
        test_predictions:
            List of predicted values on test data.
        oof_predictions:
            List of predicted values on out-of-fold training data.
        y:
            Target value
        estimator:
            Estimator used for the 2nd-level model.
            If ``None``, the default estimator (auto-tuned linear model) will be used.
        cv:
            int, cross-validation generator or an iterable which determines the cross-validation splitting strategy.

            - None, to use the default ``KFold(5, random_state=0, shuffle=True)``,
            - integer, to specify the number of folds in a ``(Stratified)KFold``,
            - CV splitter (the instance of ``BaseCrossValidator``),
            - An iterable yielding (train, test) splits as arrays of indices.
        groups:
            Group labels for the samples. Only used in conjunction with a “Group” cv instance (e.g., ``GroupKFold``).
        type_of_target:
            The type of target variable. If ``auto``, type is inferred by ``sklearn.utils.multiclass.type_of_target``.
            Otherwise, ``binary``, ``continuous``, or ``multiclass`` are supported.
        eval_func:
            Evaluation metric used for calculating result score. Used only if ``oof_predictions`` and ``y`` are given.
    Returns:
        Namedtuple with following members

        * test_prediction:
            numpy array, Average prediction on test data.
        * oof_prediction:
            numpy array, Average prediction on Out-of-Fold validation data. ``None`` if ``oof_predictions`` = ``None``.
        * score:
            float, Calculated score on Out-of-Fold data. ``None`` if ``eval_func`` is ``None``.
    """
    assert len(oof_predictions) == len(test_predictions), "Number of oof and test predictions should be same"

    def _stack(predictions):
        if predictions[0].ndim == 1:
            predictions = [p.reshape(len(p), -1) for p in predictions]
        return np.hstack(predictions)

    X_train = convert_input(_stack(oof_predictions))
    y = convert_input_vector(y, X_train.index)
    X_test = convert_input(_stack(test_predictions))

    assert len(X_train) == len(y)

    if type_of_target == 'auto':
        type_of_target = multiclass.type_of_target(y)

    if estimator is None:
        # if estimator is None, tuned linear estimator is used
        if type_of_target == 'continuous':
            estimator = Ridge(normalize=True, random_state=0)
            param_grid = {
                'alpha': [0.001, 0.01, 0.1, 1, 10],
            }
        else:
            estimator = LogisticRegression(random_state=0)
            param_grid = {
                'penalty': ['l1', 'l2'],
                'C': [0.001, 0.01, 0.1, 1, 10],
            }
        grid_search = GridSearchCV(estimator, param_grid, cv=cv)
        grid_search.fit(X_train, y, groups=groups)
        estimator = grid_search.best_estimator_

    result = cross_validate(estimator, X_train, y, X_test, cv=cv, groups=groups, eval_func=eval_func)
    score = result.scores[-1] if result.scores else None

    return EnsembleResult(result.test_prediction, result.oof_prediction, score)



# Modified work:
# -----------------------------------------------------------------------------
# Copyright (c) 2020 Kota Yuhara (@wakamezake)
# -----------------------------------------------------------------------------

# Original work of aggregation:
# https://github.com/pfnet-research/xfeat/blob/master/xfeat/helper.py
# -----------------------------------------------------------------------------
# MIT License
# 
# Copyright (c) 2020 Preferred Networks, Inc.
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -----------------------------------------------------------------------------


from types import LambdaType, FunctionType
from typing import List, Callable, Union, Tuple

import pandas as pd
from pandas.core.common import get_callable_name


def _is_lambda_function(obj):
    """
    Example:
        >>> import numpy as np
        >>> def custom_function(x): return np.sum(x)
        >>> _is_lambda_function(lambda x: np.sum(x))
        True
        >>> _is_lambda_function(np.sum)
        False
        >>> _is_lambda_function(custom_function)
        False
    """
    # It's worth noting that types.LambdaType is an alias for types.FunctionType
    return isinstance(obj, LambdaType) and obj.__name__ == "<lambda>"


def aggregation(
        input_df: pd.DataFrame,
        group_key: str,
        group_values: List[str],
        agg_methods: List[Union[str, FunctionType]],
) -> Tuple[pd.DataFrame, List[str]]:
    """
    Aggregate values after grouping table rows by a given key.

    Args:
        input_df:
            Input data frame.
        group_key:
            Used to determine the groups for the groupby.
        group_values:
            Used to aggregate values for the groupby.
        agg_methods:
            List of function or function names, e.g. ['mean', 'max', 'min', numpy.mean].
            Do not use a lambda function because the name attribute of the lambda function cannot generate a unique string of column names in <lambda>.
    Returns:
        Tuple of output dataframe and new column names.
    """
    new_df = input_df.copy()

    new_cols = []
    for agg_method in agg_methods:
        if _is_lambda_function(agg_method):
            raise ValueError('Not supported lambda function.')
        elif isinstance(agg_method, str):
            pass
        elif isinstance(agg_method, FunctionType):
            pass
        else:
            raise ValueError('Supported types are: {} or {}.'
                             ' Got {} instead.'.format(str, Callable, type(agg_method)))

    for agg_method in agg_methods:
        for col in group_values:
            # only str or FunctionType
            if isinstance(agg_method, str):
                agg_method_name = agg_method
            else:
                agg_method_name = get_callable_name(agg_method)
            new_col = "agg_{}_{}_by_{}".format(agg_method_name, col, group_key)

            df_agg = (
                input_df[[col] + [group_key]].groupby(group_key)[[col]].agg(
                    agg_method)
            )
            df_agg.columns = [new_col]
            new_cols.append(new_col)
            new_df = new_df.merge(
                df_agg, how="left", right_index=True, left_on=group_key
            )

    return new_df, new_cols



from codecs import open
from os import path

from setuptools import find_packages, setup


def get_long_description():
    here = path.abspath(path.dirname(__file__))

    with open(path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = f.read()
    return long_description


def get_version():
    version_filepath = path.join(path.dirname(__file__), 'nyaggle', 'version.py')
    with open(version_filepath) as f:
        for line in f:
            if line.startswith('__version__'):
                return line.strip().split()[-1][1:-1]


setup(
    name='nyaggle',
    packages=find_packages(),

    version=get_version(),

    license='MIT',

    install_requires=[
        'category_encoders',
        'matplotlib',
        'more-itertools',
        'numpy',
        'optuna>=1.0.0',
        'pandas',
        'pyarrow',
        'seaborn',
        'sklearn',
        'tqdm',
        'transformers>=2.3.0',
    ],

    extras_require={
        'all': ['catboost>=0.17', 'lightgbm', 'xgboost', 'torch', 'mlflow']
    },

    author='nyanp',
    author_email='Noumi.Taiga@gmail.com',
    url='https://github.com/nyanp/nyaggle',
    description='Code for Kaggle and Offline Competitions.',
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    keywords='nyaggle kaggle',
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ]
)



from collections import namedtuple
from typing import Iterable, List, Optional, Union

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.metrics import roc_auc_score
from sklearn.model_selection import KFold, BaseCrossValidator

from nyaggle.environment import requires_lightgbm
from nyaggle.util import is_instance
from nyaggle.validation.cross_validate import cross_validate
from nyaggle.validation.split import Take

ADVResult = namedtuple('ADVResult', ['auc', 'importance'])


def adversarial_validate(X_train: pd.DataFrame,
                         X_test: pd.DataFrame,
                         importance_type: str = 'gain',
                         estimator: Optional[BaseEstimator] = None,
                         categorical_feature: List[str] = None,
                         cv: Optional[Union[int, Iterable, BaseCrossValidator]] = None) -> ADVResult:
    """
    Perform adversarial validation between X_train and X_test.

    Args:
        X_train:
            Training data
        X_test:
            Test data
        importance_type:
            The type of feature importance calculated.
        estimator:
            The custom estimator. If None, LGBMClassifier is automatically used.
            Only LGBMModel or CatBoost instances are supported.
        categorical_feature:
            List of categorical column names. If ``None``, categorical columns are automatically determined by dtype.
        cv:
            Cross validation split. If ``None``, the first fold out of 5 fold is used as validation.
    Returns:
        Namedtuple with following members

        * auc:
            float, ROC AUC score of adversarial validation.
        * importance:
            pandas DataFrame, feature importance of adversarial model (order by importance)

    Example:
        >>> from sklearn.model_selection import train_test_split
        >>> from nyaggle.testing import make_regression_df
        >>> from nyaggle.validation import adversarial_validate

        >>> X, y = make_regression_df(n_samples=8)
        >>> X_train, X_test, y_train, y_test = train_test_split(X, y)
        >>> auc, importance = cross_validate(X_train, X_test)
        >>>
        >>> print(auc)
        0.51078231
        >>> importance.head()
        feature importance
        col_1   231.5827204
        col_5   207.1837266
        col_7   188.6920685
        col_4   174.5668498
        col_9   170.6438643
    """
    concat = pd.concat([X_train, X_test]).copy().reset_index(drop=True)
    y = np.array([1] * len(X_train) + [0] * len(X_test))

    if estimator is None:
        requires_lightgbm()
        from lightgbm import LGBMClassifier
        estimator = LGBMClassifier(n_estimators=10000, objective='binary', importance_type=importance_type,
                                   random_state=0)
    else:
        assert is_instance(estimator, ('lightgbm.sklearn.LGBMModel', 'catboost.core.CatBoost')), \
            'Only CatBoostClassifier or LGBMClassifier is allowed'

    if cv is None:
        cv = Take(1, KFold(5, shuffle=True, random_state=0))

    fit_params = {'verbose': -1}
    if categorical_feature:
        fit_params['categorical_feature'] = categorical_feature

    result = cross_validate(estimator, concat, y, None, cv=cv,
                            eval_func=roc_auc_score, fit_params=fit_params, importance_type=importance_type)

    importance = pd.concat(result.importance)
    importance = importance.groupby('feature')['importance'].mean().reset_index()
    importance.sort_values(by='importance', ascending=False, inplace=True)
    importance.reset_index(drop=True, inplace=True)

    return ADVResult(result.scores[-1], importance)



# pytorch

try:
    import torch
    _has_torch = True
except ImportError:
    _has_torch = False


def requires_torch():
    if not _has_torch:
        raise ImportError('You need to install pytorch before using this API.')


# mlflow

try:
    import mlflow
    _has_mlflow = True
except ImportError:
    _has_mlflow = False


def requires_mlflow():
    if not _has_mlflow:
        raise ImportError('You need to install mlflow before using this API.')


# lightgbm


try:
    import lightgbm
    _has_lightgbm = True
except ImportError:
    _has_lightgbm = False


def requires_lightgbm():
    if not _has_lightgbm:
        raise ImportError('You need to install lightgbm before using this API.')


# lightgbm


try:
    import catboost
    _has_catboost = True
    # TODO check catboost version >= 0.17
except ImportError:
    _has_catboost = False


def requires_catboost():
    if not _has_catboost:
        raise ImportError('You need to install catboost before using this API.')


# xgboost


try:
    import xgboost
    _has_xgboost = True
except ImportError:
    _has_xgboost = False


def requires_xgboost():
    if not _has_xgboost:
        raise ImportError('You need to install xgboost before using this API.')
